from rich.style import Style
from rich.text import Text
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Container, Horizontal, ScrollableContainer
from textual.screen import Screen
from textual.widgets import (
    Button,
    DataTable,
    Label,
    Rule,
    Static,
    TabbedContent,
    TabPane,
    TextArea,
)
from textual.widgets.text_area import TextAreaTheme

from dolphie.Modules.Functions import format_number
from dolphie.Widgets.TopBar import TopBar


class ThreadScreen(Screen):
    CSS = """
        ThreadScreen {
            background: #0a0e1b;

            & #explain_table {
                margin-top: 1;
                background: #101626;
                border: tall #1d253e;
                overflow-x: auto;
                min-height: 5;
                max-height: 15;
                width: 100%;
            }

            & #explain_failure {
                margin-top: 1;
                max-width: 120;
            }

            & Container {
                height: auto;
            }

            & #thread_container {
                margin-top: 1;
                height: auto;
                layout: horizontal;
            }

            & .title {
                width: 100%;
                content-align: center middle;
                color: #bbc8e8;
                text-style: bold;
            }

            & Center {
                height: auto;
            }

            & #query {
                width: auto;
            }

            & .container > Center {
                layout: horizontal;
            }

            & ScrollableContainer {
                height: auto;
                width: 50vw;
                max-height: 15;
            }

            & .table {
                content-align: center middle;
                background: #0f1525;
                border: tall #1d253e;
                padding-left: 1;
                padding-right: 1;
                height: auto;
            }

            & TextArea {
                border: tall #1d253e;
                max-height: 25;
            }

            & .copy-button {
                margin-left: 1;
                margin-right: 1;
                background: #1d253e;
                color: #bbc8e8;
                border: tall #2F3C59;
            }

            & .copy-button:hover {
                background: #2F3C59;
                color: #ffffff;
            }

            & .copy-buttons {
                margin-bottom: 1;
            }
        }

    """

    BINDINGS = [
        Binding("q", "app.pop_screen", "", show=False),
        Binding("c", "copy_query", "Copy Query", show=True),
    ]

    def __init__(
        self,
        connection_status: str,
        app_version: str,
        host: str,
        thread_table: str,
        user_thread_attributes_table: str,
        query: str,
        explain_data: str,
        explain_json_data: str,
        explain_failure: str,
        transaction_history_table: str,
    ):
        super().__init__()

        self.connection_status = connection_status
        self.app_version = app_version
        self.host = host

        self.thread_table = thread_table
        self.user_thread_attributes_table = user_thread_attributes_table
        self.formatted_query = query
        self.explain_data = explain_data
        self.explain_json_data = explain_json_data
        self.explain_failure = explain_failure
        self.transaction_history_table = transaction_history_table

        dracula = TextAreaTheme.get_builtin_theme("dracula")
        dracula.base_style = Style(bgcolor="#101626")
        dracula.gutter_style = Style(color="#606e88")
        dracula.cursor_line_gutter_style = Style(color="#95a7c7", bgcolor="#20243b")
        dracula.cursor_line_style = Style(bgcolor="#20243b")
        dracula.selection_style = Style(bgcolor="#293c71")
        dracula.cursor_style = Style(bgcolor="#7a8ab2", color="#121e3a")
        dracula.syntax_styles = {
            "json.label": Style(color="#879bca", bold=True),
            "number": Style(color="#ca87a5"),
        }

        self.explain_json_text_area = TextArea(theme="dracula", show_line_numbers=True, read_only=True)

    def copy_to_clipboard(self, text: str, content_type: str = "content"):
        """Copy text to clipboard and show notification."""
        try:
            self.app.copy_to_clipboard(text)
            self.notify(f"Copied {content_type} to clipboard!", severity="information")
        except Exception as e:
            self.notify(f"Failed to copy {content_type} to clipboard: {e}", severity="error")

    def action_copy_query(self) -> None:
        """Action to copy the query via keyboard shortcut."""
        if self.formatted_query:
            self.copy_to_clipboard(self.formatted_query.code, "query")
        else:
            self.notify("No query to copy", severity="warning")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button press events."""
        if event.button.id == "copy_query_btn":
            if self.formatted_query:
                self.copy_to_clipboard(self.formatted_query.code, "query")
            else:
                self.notify("No query to copy", severity="warning")
        elif event.button.id == "copy_json_btn":
            if self.explain_json_data:
                self.copy_to_clipboard(self.explain_json_data, "JSON data")
            else:
                self.notify("No JSON data to copy", severity="warning")

    def on_mount(self):
        self.query_one("#thread_table").update(self.thread_table)
        self.query_one("#query").update(self.formatted_query)

        if self.transaction_history_table:
            self.query_one("#transaction_history_table").update(self.transaction_history_table)
        else:
            self.query_one("#transaction_history_container").display = False

        if self.user_thread_attributes_table:
            self.query_one("#user_thread_attributes_table").update(self.user_thread_attributes_table)
        else:
            self.query_one("#user_thread_attributes_table").display = False

        if self.formatted_query:
            if self.explain_failure:
                self.query_one("#explain_tabbed_content").display = False
                self.query_one("#explain_failure").update(Text.from_markup(self.explain_failure))
            elif self.explain_data:
                self.query_one("#explain_failure").display = False

                explain_table = self.query_one("#explain_table", DataTable)

                columns = []
                for row in self.explain_data:
                    values = []
                    for column, value in row.items():
                        # Exclude possbile_keys field since it takes up too much space
                        if column == "possible_keys":
                            continue

                        # Don't duplicate columns
                        if column not in columns:
                            explain_table.add_column(f"[label]{column}")
                            columns.append(column)

                        if column == "key" and value is None:
                            value = "[b white on #B30000]NO INDEX[/b white on #B30000]"

                        if column == "rows":
                            value = format_number(value)

                        values.append(str(value))

                    explain_table.add_row(*values)
            else:
                self.query_one("#explain_table").display = False
                self.query_one("#explain_failure").display = False
        else:
            self.query_one("#query_container").display = False

        if self.explain_json_data:
            self.explain_json_text_area.text = self.explain_json_data
        else:
            self.query_one("#explain_tabbed_content").display = False

    def compose(self) -> ComposeResult:
        yield TopBar(connection_status=self.connection_status, app_version=self.app_version, host=self.host)

        with Container(id="thread_container", classes="container"):
            with Container():
                yield Label("Thread Details", classes="title")
                yield ScrollableContainer(Static(id="thread_table"), classes="table")
            with Container():
                yield Label("Thread Attributes", classes="title")
                yield ScrollableContainer(Static(id="user_thread_attributes_table"), classes="table")

        with Container(id="query_container", classes="container"):
            yield Rule(line_style="heavy")

            # Only show copy buttons if there's data to copy
            copy_buttons = []
            if self.formatted_query:
                copy_buttons.append(Button("📋 Copy Query", id="copy_query_btn", classes="copy-button"))
            if self.explain_json_data:
                copy_buttons.append(Button("📋 Copy JSON", id="copy_json_btn", classes="copy-button"))

            # Only create the horizontal container if there are buttons to show
            if copy_buttons:
                with Horizontal(classes="button_container copy-buttons"):
                    yield from copy_buttons

            yield Label("Query", classes="title")
            yield Center(Static(id="query", shrink=True, classes="table"))

            yield Center(Label("", id="explain_failure"))
            with TabbedContent(id="explain_tabbed_content", classes="container"):
                with TabPane("Table", id="table_explain_tab", classes="container"):
                    yield DataTable(show_cursor=False, id="explain_table", classes="table")

                with TabPane("JSON", id="json_explain_tab", classes="container"):
                    yield Center(self.explain_json_text_area)

        with Container(id="transaction_history_container", classes="container"):
            yield Rule(line_style="heavy")
            yield Label("Transaction History", id="transaction_history_label", classes="title")
            yield Center(
                Static(id="transaction_history_table", shrink=True, classes="table"),
                id="transaction_history_table_center",
            )



from datetime import datetime, timedelta

from dolphie.Modules.Functions import format_bytes, format_number
from dolphie.Modules.MetricManager import MetricData
from dolphie.Modules.MySQL import ConnectionSource
from dolphie.Modules.TabManager import Tab
from dolphie.Panels import Replication as ReplicationPanel
from rich.style import Style
from rich.table import Table


def create_panel(tab: Tab) -> Table:
    dolphie = tab.dolphie

    global_status = dolphie.global_status
    global_variables = dolphie.global_variables
    binlog_status = dolphie.binlog_status

    table_title_style = Style(color="#bbc8e8", bold=True)

    ####################
    # Host Information #
    ####################
    table_information = Table(
        show_header=False,
        box=None,
        title=f"{dolphie.panels.dashboard.formatted_key}Host Information",
        title_style=table_title_style,
    )

    if dolphie.replicaset:
        host_type = "InnoDB ReplicaSet"
    elif dolphie.innodb_cluster_read_replica:
        host_type = "InnoDB Cluster Read Replica"
    elif dolphie.innodb_cluster:
        host_type = "InnoDB Cluster"
    elif dolphie.group_replication:
        host_type = "Group Replication"
    elif dolphie.galera_cluster:
        host_type = "Galera Cluster"
    else:
        host_type = "MariaDB" if dolphie.connection_source_alt == ConnectionSource.mariadb else "MySQL"

    replicas = len(dolphie.replica_manager.available_replicas) if dolphie.replica_manager.available_replicas else 0

    table_information.add_column()
    table_information.add_column(min_width=25, max_width=35)
    table_information.add_row("[label]Version", f"{dolphie.host_distro} {dolphie.host_version}")
    if global_variables.get("version_compile_os") and global_variables.get("version_compile_machine"):
        table_information.add_row(
            "[label]",
            f"{global_variables['version_compile_os']} ({global_variables['version_compile_machine']})",
        )
    table_information.add_row("[label]Type", host_type)

    if dolphie.galera_cluster or dolphie.group_replication or dolphie.innodb_cluster:
        if dolphie.galera_cluster:
            nodes = dolphie.global_status.get("wsrep_cluster_size", 0)
            node_state = dolphie.global_status.get("wsrep_local_state_comment", "N/A")
            if node_state == "Synced":
                state_colored = f"[green]{node_state}[/green]"
            elif node_state in ("Donor/Desynced", "Donor"):
                state_colored = f"[yellow]{node_state}[/yellow]"
            else:
                state_colored = f"[red]{node_state}[/red]"
            node_value = f"{nodes} ({state_colored})"
        else:
            nodes = len(dolphie.group_replication_members)
            node_value = str(nodes)
        if replicas:
            node_value += f" [label]Replicas[/label] {replicas}"
        table_information.add_row("[label]Nodes", node_value)
    else:
        table_information.add_row("[label]Replicas", str(replicas))

    table_information.add_row(
        "[label]Threads",
        f"[label]con[/label] {format_number(global_status['Threads_connected'])}"
        f"[highlight]/[/highlight][label]run[/label] {format_number(global_status['Threads_running'])}"
        f"[highlight]/[/highlight][label]cac[/label] {format_number(global_status['Threads_cached'])}",
    )
    table_information.add_row(
        "[label]Tables",
        f"[label]open[/label] {format_number(global_status['Open_tables'])}"
        f"[highlight]/[/highlight][label]opened[/label] {format_number(global_status['Opened_tables'])}",
    )
    table_information.add_row("[label]Uptime", str(timedelta(seconds=global_status["Uptime"])))

    if not dolphie.replay_file:
        runtime = str(datetime.now().astimezone() - dolphie.dolphie_start_time).split(".")[0]
        table_information.add_row(
            "[label]Runtime",
            f"{runtime} [label]Latency[/label] {round(dolphie.worker_processing_time, 2)}s",
        )
    else:
        if dolphie.worker_processing_time:
            table_information.add_row("[label]Latency", f"{round(dolphie.worker_processing_time, 2)}s")

    tab.dashboard_section_1.update(table_information)

    ######################
    # System Utilization #
    ######################
    table = create_system_utilization_table(tab)

    if table:
        tab.dashboard_section_6.update(table)

    ###########
    # InnoDB  #
    ###########
    table_innodb = Table(show_header=False, box=None, title="InnoDB", title_style=table_title_style)

    table_innodb.add_column()
    table_innodb.add_column(width=9)

    # Calculate InnoDB memory read hit efficiency
    ib_pool_disk_reads = global_status.get("Innodb_buffer_pool_reads", 0)
    ib_pool_mem_reads = global_status.get(
        "Innodb_buffer_pool_read_requests", 1
    )  # Default to 1 to avoid division by zero

    if ib_pool_disk_reads >= ib_pool_mem_reads:
        innodb_efficiency = "[red]0.00%"
    else:
        efficiency = 100 - (ib_pool_disk_reads / ib_pool_mem_reads * 100)

        if efficiency > 90:
            color_code = "green"
        elif efficiency > 80:
            color_code = "yellow"
        else:
            color_code = "red"

        innodb_efficiency = f"[{color_code}]{efficiency:.2f}%"

    # Add data to our table
    table_innodb.add_row("[label]Read Hit", innodb_efficiency)
    table_innodb.add_row(
        "[label]Chkpt Age",
        dolphie.metric_manager.get_formatted_checkpoint_age(),
    )
    table_innodb.add_row("[label]AHI Hit", dolphie.metric_manager.get_formatted_ahi_status())

    bp_instances = global_variables.get("innodb_buffer_pool_instances", 1)
    plural = "s" if bp_instances > 1 else ""
    table_innodb.add_row(f"[label]BP Instance{plural}", format_number(bp_instances))

    table_innodb.add_row("[label]BP Size", format_bytes(global_variables["innodb_buffer_pool_size"]))
    table_innodb.add_row(
        "[label]BP Available",
        format_bytes(global_variables["innodb_buffer_pool_size"] - global_status["Innodb_buffer_pool_bytes_data"]),
    )
    table_innodb.add_row("[label]BP Dirty", format_bytes(global_status["Innodb_buffer_pool_bytes_dirty"]))
    table_innodb.add_row(
        "[label]History List",
        format_number(dolphie.innodb_metrics.get("trx_rseg_history_len", "N/A")),
    )

    tab.dashboard_section_2.update(table_innodb)

    ##############
    # Binary Log #
    ##############
    table_primary = Table(show_header=False, box=None, title="Binary Log", title_style=table_title_style)

    if global_variables.get("log_bin") == "OFF" or not binlog_status or not binlog_status.get("File"):
        tab.dashboard_section_3.display = False
    else:
        tab.dashboard_section_3.display = True
        table_primary.add_column()
        table_primary.add_column(max_width=40)

        binlog_cache = 100
        binlog_cache_disk = global_status["Binlog_cache_disk_use"]
        binlog_cache_mem = global_status["Binlog_cache_use"]
        if binlog_cache_disk and binlog_cache_mem:
            if binlog_cache_disk >= binlog_cache_mem:
                binlog_cache = 0
            else:
                binlog_cache = round(100 - (binlog_cache_disk / binlog_cache_mem * 100), 2)

        table_primary.add_row("[label]File name", binlog_status["File"])
        table_primary.add_row(
            "[label]Position",
            str(binlog_status["Position"]),
        )
        table_primary.add_row(
            "[label]Size",
            format_bytes(binlog_status["Position"]),
        )
        table_primary.add_row("[label]Diff", format_bytes(binlog_status["Diff_Position"]))
        table_primary.add_row("[label]Cache Hit", f"{binlog_cache}%")

        binlog_format = global_variables.get("binlog_format", "N/A")
        if binlog_format == "ROW":
            binlog_row_image = global_variables.get("binlog_row_image", "N/A")
            table_primary.add_row("[label]Format", f"{binlog_format} ({binlog_row_image})")
        else:
            table_primary.add_row("[label]Format", binlog_format)

        if dolphie.connection_source_alt == ConnectionSource.mariadb:
            table_primary.add_row("[label]GTID Strict", global_variables.get("gtid_strict_mode", "N/A"))
            table_primary.add_row("[label]Encrypt", global_variables.get("encrypt_binlog", "N/A"))
        else:
            table_primary.add_row("[label]GTID", global_variables.get("gtid_mode", "N/A"))
            table_primary.add_row(
                "[label]Compression",
                global_variables.get("binlog_transaction_compression", "N/A"),
            )

        tab.dashboard_section_3.update(table_primary)

    ###############
    # Replication #
    ###############
    if dolphie.replication_status and not dolphie.panels.replication.visible:
        tab.dashboard_section_5.display = True
        tab.dashboard_section_5.update(ReplicationPanel.create_replication_table(tab, dashboard_table=True))
    else:
        tab.dashboard_section_5.display = False
    ###############
    # Statistics #
    ###############
    table_stats = Table(show_header=False, box=None, title="Statistics/s", title_style=table_title_style)

    table_stats.add_column()
    table_stats.add_column(min_width=6)

    # Add DML statistics
    metrics = dolphie.metric_manager.metrics.dml
    metric_labels = {
        "Queries": "Queries",
        "SELECT": "Com_select",
        "INSERT": "Com_insert",
        "UPDATE": "Com_update",
        "DELETE": "Com_delete",
        "REPLACE": "Com_replace",
        "COMMIT": "Com_commit",
        "ROLLBACK": "Com_rollback",
    }

    for label, metric_name in metric_labels.items():
        metric_data: MetricData = getattr(metrics, metric_name)

        if metric_data.values:
            table_stats.add_row(f"[label]{label}", format_number(metric_data.values[-1]))
        else:
            table_stats.add_row(f"[label]{label}", "0")

    tab.dashboard_section_4.update(table_stats)


def create_system_utilization_table(tab: Tab) -> Table:
    dolphie = tab.dolphie

    system_utilization = dolphie.system_utilization
    if not system_utilization:
        return None

    table = Table(
        show_header=False,
        box=None,
        title="System Utilization",
        title_style=Style(color="#bbc8e8", bold=True),
    )
    table.add_column()
    table.add_column(min_width=18, max_width=25)

    def format_percent(value, thresholds=(80, 90), colors=("green", "yellow", "red")):
        if value > thresholds[1]:
            return f"[{colors[2]}]{value}%[/{colors[2]}]"
        elif value > thresholds[0]:
            return f"[{colors[1]}]{value}%[/{colors[1]}]"
        return f"[{colors[0]}]{value}%[/{colors[0]}]"

    # Uptime
    uptime = system_utilization.get("Uptime", "N/A")
    table.add_row("[label]Uptime", str(timedelta(seconds=uptime)) if uptime != "N/A" else "N/A")

    # CPU
    cpu_percent_values = dolphie.metric_manager.metrics.system_cpu.CPU_Percent.values
    if cpu_percent_values:
        cpu_percent = round(cpu_percent_values[-1], 2)
        formatted_cpu_percent = format_percent(cpu_percent)
        cpu_cores = system_utilization.get("CPU_Count", "N/A")
        table.add_row("[label]CPU", f"{formatted_cpu_percent} [label]cores[/label] {cpu_cores}")
    else:
        table.add_row("[label]CPU", "N/A")

    # CPU Load
    load_averages = system_utilization.get("CPU_Load_Avg")
    if load_averages:
        formatted_load = " ".join(f"{avg:.2f}" for avg in load_averages)
        table.add_row("[label]Load", formatted_load)

    # Memory
    memory_used = dolphie.metric_manager.metrics.system_memory.Memory_Used.last_value
    memory_total = dolphie.metric_manager.metrics.system_memory.Memory_Total.last_value
    if memory_used and memory_total:
        memory_percent_used = round((memory_used / memory_total) * 100, 2)
        formatted_memory_percent_used = format_percent(memory_percent_used)
        table.add_row(
            "[label]Memory",
            (
                f"{formatted_memory_percent_used}\n{format_bytes(memory_used)}"
                f"[dark_gray]/[/dark_gray]{format_bytes(memory_total)}"
            ),
        )
    else:
        table.add_row("[label]Memory", "N/A\n")

    # Swap
    swap_used = format_bytes(system_utilization.get("Swap_Used", "N/A"))
    swap_total = format_bytes(system_utilization.get("Swap_Total", "N/A"))
    table.add_row("[label]Swap", f"{swap_used}[dark_gray]/[/dark_gray]{swap_total}")

    # Disk I/O
    disk_read_values = dolphie.metric_manager.metrics.system_disk_io.Disk_Read.values
    disk_write_values = dolphie.metric_manager.metrics.system_disk_io.Disk_Write.values
    if disk_read_values and disk_write_values:
        last_disk_read = format_number(disk_read_values[-1])
        last_disk_write = format_number(disk_write_values[-1])
        table.add_row(
            "[label]Disk",
            f"[label]IOPS R[/label] {last_disk_read}\n[label]IOPS W[/label] {last_disk_write}",
        )
    else:
        table.add_row("[label]Disk", "[label]IOPS R[/label] N/A\n[label]IOPS W[/label] N/A")

    return table



from dataclasses import dataclass


@dataclass
class ProxySQLQueries:
    processlist: str = """
        SELECT
            SessionID      AS id,
            user           AS user,
            db             AS db,
            cli_host       AS frontend_host,
            hostgroup      AS hostgroup,
            srv_host       AS backend_host,
            command        AS command,
            time_ms        AS time,
            info           AS query,
            extended_info  AS extended_info
        FROM
            stats_mysql_processlist
        WHERE
            $1
    """
    mysql_stats: str = """
        SELECT
            Variable_Name  AS Variable_name,
            Variable_Value AS Value
        FROM
            stats_mysql_global
    """
    connection_pool_data: str = """
        SELECT
            SUM(ConnUsed) AS connection_pool_connections
        FROM
            stats_mysql_connection_pool
    """
    user_stats: str = """
        SELECT DISTINCT
            su.username,
            frontend_connections,
            frontend_max_connections,
            default_hostgroup,
            default_schema,
            use_ssl
        FROM
            stats_mysql_users su JOIN
            runtime_mysql_users ru ON su.username = ru.username
        WHERE
            frontend_connections > 0
        ORDER BY
            frontend_connections DESC
    """
    hostgroup_summary: str = """
        SELECT
            *
        FROM
            stats_mysql_connection_pool LEFT JOIN
            runtime_mysql_servers ON hostgroup = hostgroup_id AND srv_host = hostname AND srv_port = port
        ORDER BY
            hostgroup
    """
    query_rules_summary: str = """
        SELECT
            *,
            hits AS hits_s
        FROM
            stats_mysql_query_rules
            JOIN runtime_mysql_query_rules USING (rule_id)
        WHERE
            active = 1
        ORDER BY
            hits DESC
    """
    command_stats: str = """
        SELECT
            *,
            Total_cnt AS Total_cnt_s
        FROM
            stats_mysql_commands_counters
        WHERE
            Total_cnt != 0
        ORDER BY
            Total_cnt DESC
    """
    memory_metrics: str = """
        SELECT
            *
        FROM
            stats_memory_metrics
        ORDER BY
            CAST(Variable_Value AS DECIMAL) DESC;
    """
    query_errors: str = """
        SELECT
            *
        FROM
            stats_mysql_errors
        ORDER BY
            count_star DESC
    """
    variables: str = "SHOW GLOBAL VARIABLES"


@dataclass
class MySQLQueries:
    pl_query: str = """
        SELECT
            id,
            User                    AS user,
            Host                    AS host,
            db                      AS db,
            Command                 AS command,
            Time                    AS time,
            Info                    AS query,
            State                   AS state,
            trx_query               AS trx_query,
            trx_state               AS trx_state,
            trx_operation_state     AS trx_operation_state,
            trx_rows_locked         AS trx_rows_locked,
            trx_rows_modified       AS trx_rows_modified,
            trx_concurrency_tickets AS trx_concurrency_tickets,
            ""                      AS connection_type,
            TIMESTAMPDIFF(SECOND, trx_started, NOW()) AS trx_time
        FROM
            information_schema.PROCESSLIST pl
            LEFT JOIN information_schema.innodb_trx ON trx_mysql_thread_id = pl.Id
        WHERE
            command != 'Daemon'
            $1
    """
    ps_query: str = """
        SELECT
            processlist_id          AS id,
            thread_id               AS mysql_thread_id,
            processlist_user        AS user,
            processlist_host        AS host,
            processlist_db          AS db,
            processlist_command     AS command,
            processlist_time        AS time,
            processlist_info        AS query,
            processlist_state       AS state,
            trx_query               AS trx_query,
            trx_state               AS trx_state,
            trx_operation_state     AS trx_operation_state,
            trx_rows_locked         AS trx_rows_locked,
            trx_rows_modified       AS trx_rows_modified,
            trx_concurrency_tickets AS trx_concurrency_tickets,
            connection_type         AS connection_type,
            TIMESTAMPDIFF(SECOND, trx_started, NOW()) AS trx_time
        FROM
            performance_schema.threads t
            LEFT JOIN information_schema.innodb_trx tx ON trx_mysql_thread_id = t.processlist_id
        WHERE
            processlist_id IS NOT NULL AND
            processlist_time IS NOT NULL AND
            processlist_command != 'Daemon'
            $1
    """
    locks_query: str = """
        SELECT /*+ MAX_EXECUTION_TIME(10000) */
            wait_age,
            locked_type,
            waiting_pid,
            waiting_trx_age,
            waiting_trx_rows_modified,
            waiting_trx_rows_locked,
            waiting_lock_mode,
            waiting_query
            blocking_pid,
            blocking_trx_age,
            blocking_trx_rows_modified,
            blocking_trx_rows_locked,
            blocking_lock_mode,
            blocking_query
        FROM
            sys.innodb_lock_waits
    """
    ps_replica_lag: str = """
        SELECT MAX(`lag`) AS Seconds_Behind_Master
            FROM (
                SELECT MAX(TIMESTAMPDIFF(SECOND, APPLYING_TRANSACTION_IMMEDIATE_COMMIT_TIMESTAMP, NOW())) AS `lag`
                FROM performance_schema.replication_applier_status_by_worker

                UNION

                SELECT MIN(
                    CASE
                        WHEN
                            LAST_QUEUED_TRANSACTION = 'ANONYMOUS' OR
                            LAST_APPLIED_TRANSACTION = 'ANONYMOUS' OR
                            GTID_SUBTRACT(LAST_QUEUED_TRANSACTION, LAST_APPLIED_TRANSACTION) = ''
                            THEN 0
                        ELSE
                            TIMESTAMPDIFF(SECOND, LAST_APPLIED_TRANSACTION_IMMEDIATE_COMMIT_TIMESTAMP, NOW())
                    END
                ) AS `lag`
                FROM performance_schema.replication_applier_status_by_worker w
                JOIN performance_schema.replication_connection_status s ON s.channel_name = w.channel_name
            ) required
    """
    ps_disk_io: str = """
        SELECT
            CONVERT(SUM(SUM_NUMBER_OF_BYTES_READ), UNSIGNED) AS io_read,
            CONVERT(SUM(SUM_NUMBER_OF_BYTES_WRITE), UNSIGNED) AS io_write
        FROM
            `performance_schema`.`file_summary_by_event_name`
        WHERE
            `file_summary_by_event_name`.`EVENT_NAME` LIKE 'wait/io/file/%' AND
            `file_summary_by_event_name`.`COUNT_STAR` > 0
    """
    heartbeat_replica_lag: str = """
        SELECT
            TIMESTAMPDIFF(SECOND, MAX(ts), NOW()) AS Seconds_Behind_Master
        FROM
            $1
    """
    ps_find_replicas: str = """
        SELECT
            t.PROCESSLIST_ID AS id,
            t.PROCESSLIST_USER AS user,
            t.PROCESSLIST_HOST AS host,
            CONVERT (
                CAST( CONVERT ( uvt.VARIABLE_VALUE USING latin1 ) AS BINARY ) USING utf8
            ) AS replica_uuid
        FROM
            `performance_schema`.threads AS t JOIN
            `performance_schema`.user_variables_by_thread AS uvt ON t.THREAD_ID = uvt.THREAD_ID
        WHERE
            t.PROCESSLIST_COMMAND LIKE 'Binlog Dump%'
            AND uvt.VARIABLE_NAME = 'slave_uuid'
    """
    pl_find_replicas: str = """
        SELECT
            Id   AS id,
            User AS user,
            Host AS host,
            ''   AS replica_uuid
        FROM
            information_schema.PROCESSLIST
        WHERE
            Command Like 'Binlog Dump%'
    """
    mariadb_find_replicas: str = """
        SELECT
            PROCESSLIST_ID AS id,
            PROCESSLIST_USER AS user,
            PROCESSLIST_HOST AS host
        FROM
            `performance_schema`.threads
        WHERE
            PROCESSLIST_COMMAND LIKE 'Binlog Dump%'
    """
    ps_user_statisitics: str = """
        SELECT
            u.user AS user,
            total_connections AS total_connections,
            current_connections AS current_connections,
            CONVERT(SUM(sum_rows_affected), UNSIGNED) AS rows_affected,
            CONVERT(SUM(sum_rows_sent), UNSIGNED) AS rows_sent,
            CONVERT(SUM(sum_rows_examined), UNSIGNED) AS rows_examined,
            CONVERT(SUM(sum_created_tmp_disk_tables), UNSIGNED) AS created_tmp_disk_tables,
            CONVERT(SUM(sum_created_tmp_tables), UNSIGNED) AS created_tmp_tables,
            plugin AS plugin
        FROM
            performance_schema.users u
            JOIN performance_schema.events_statements_summary_by_user_by_event_name ess ON u.user = ess.user
            JOIN mysql.user mysql_user ON mysql_user.user = u.user
        WHERE
            current_connections != 0
        GROUP BY
            user
        ORDER BY
            current_connections DESC
    """
    ps_user_statisitics_56: str = """
        SELECT
            u.user AS user,
            total_connections,
            current_connections,
            CONVERT(SUM(sum_rows_affected), UNSIGNED) AS rows_affected,
            CONVERT(SUM(sum_rows_sent), UNSIGNED) AS rows_sent,
            CONVERT(SUM(sum_rows_examined), UNSIGNED) AS rows_examined,
            CONVERT(SUM(sum_created_tmp_disk_tables), UNSIGNED) AS created_tmp_disk_tables,
            CONVERT(SUM(sum_created_tmp_tables), UNSIGNED) AS created_tmp_tables,
            plugin
        FROM
            performance_schema.users u
            JOIN performance_schema.events_statements_summary_by_user_by_event_name ess ON u.user = ess.user
            JOIN mysql.user mysql_user ON mysql_user.user = u.user
        WHERE
            current_connections != 0
        GROUP BY
            user
        ORDER BY
            current_connections DESC
    """
    user_thread_attributes: str = """
        SELECT
            ATTR_NAME,
            ATTR_VALUE
        FROM
            `performance_schema`.session_connect_attrs sca
            JOIN `performance_schema`.threads t ON sca.PROCESSLIST_ID = t.processlist_id
        WHERE
            t.processlist_id = $1
        ORDER BY
            ATTR_NAME
    """
    ddls: str = """
        SELECT
            t.processlist_id,
            stmt.sql_text AS sql_text,
            stage.event_name AS state,
            CONCAT(ROUND(100 * stage.work_completed / stage.work_estimated, 2), "%") AS percentage_completed,
            stmt.timer_wait AS started_ago,
            CONVERT(stmt.timer_wait / ROUND(100 * stage.work_completed / stage.work_estimated, 2) * 100,
                UNSIGNED) AS estimated_full_time,
            CONVERT((stmt.timer_wait / ROUND(100 * stage.work_completed / stage.work_estimated, 2) * 100)
                - stmt.timer_wait, UNSIGNED) AS estimated_remaining_time,
            CONVERT(SUM(`mt`.`CURRENT_NUMBER_OF_BYTES_USED`), UNSIGNED) AS memory
        FROM
            `performance_schema`.`events_statements_current` stmt JOIN
            `performance_schema`.`events_stages_current` stage ON stage.nesting_event_id = stmt.event_id JOIN
            `performance_schema`.`memory_summary_by_thread_by_event_name` `mt` ON `mt`.thread_id = stmt.thread_id JOIN
            `performance_schema`.`threads` t ON t.thread_id = stmt.thread_id
        WHERE
            stage.event_name LIKE 'stage/innodb/alter%'
        GROUP BY
            t.processlist_id
    """
    file_summary_by_instance: str = """
        SELECT
            FILE_NAME,
            EVENT_NAME,
            SUM_TIMER_WAIT,
            COUNT_READ,
            COUNT_WRITE,
            COUNT_MISC,
            SUM_NUMBER_OF_BYTES_READ,
            SUM_NUMBER_OF_BYTES_WRITE
        FROM
            performance_schema.file_summary_by_instance
        WHERE
            COUNT_STAR > 0
    """
    table_io_waits_summary_by_table: str = """
        SELECT
            CONCAT(OBJECT_SCHEMA,'.', OBJECT_NAME) AS OBJECT_TABLE,
            COUNT_STAR,
            SUM_TIMER_WAIT,
            COUNT_FETCH,
            SUM_TIMER_FETCH,
            COUNT_INSERT,
            SUM_TIMER_INSERT,
            COUNT_UPDATE,
            SUM_TIMER_UPDATE,
            COUNT_DELETE,
            SUM_TIMER_DELETE
        FROM
            performance_schema.table_io_waits_summary_by_table
        WHERE
            COUNT_STAR > 0
    """
    table_statements_summary_by_digest: str = """
        SELECT
            `digest`,
            `digest_text`,
            `schema_name`,
            CONVERT(SUM(`sum_no_good_index_used`), UNSIGNED) AS sum_no_good_index_used,
            CONVERT(SUM(`sum_no_index_used`), UNSIGNED) AS sum_no_index_used,
            CONVERT(SUM(`count_star`), UNSIGNED) AS count_star,
            CONVERT(SUM(`sum_errors`), UNSIGNED) AS sum_errors,
            CONVERT(SUM(`sum_warnings`), UNSIGNED) AS sum_warnings,
            CONVERT(SUM(`sum_timer_wait`), UNSIGNED) AS sum_timer_wait,
            CONVERT(SUM(`sum_lock_time`), UNSIGNED) AS sum_lock_time,
            CONVERT(SUM(`sum_rows_sent`), UNSIGNED) AS sum_rows_sent,
            CONVERT(SUM(`sum_rows_examined`), UNSIGNED) AS sum_rows_examined,
            CONVERT(SUM(`sum_rows_affected`), UNSIGNED) AS sum_rows_affected
        FROM
            `performance_schema`.`events_statements_summary_by_digest`
        GROUP BY
            `digest`
        ORDER BY
            SUM(`sum_timer_wait`) DESC,
            MAX(`last_seen`) DESC;
    """
    table_statements_summary_by_digest_80: str = """
        SELECT
            `digest`,
            `digest_text`,
            `schema_name`,
            `query_sample_text`,
            CONVERT(SUM(`sum_no_good_index_used`), UNSIGNED) AS sum_no_good_index_used,
            CONVERT(SUM(`sum_no_index_used`), UNSIGNED) AS sum_no_index_used,
            CONVERT(SUM(`count_star`), UNSIGNED) AS count_star,
            CONVERT(SUM(`sum_errors`), UNSIGNED) AS sum_errors,
            CONVERT(SUM(`sum_warnings`), UNSIGNED) AS sum_warnings,
            CONVERT(SUM(`sum_timer_wait`), UNSIGNED) AS sum_timer_wait,
            CONVERT(SUM(`sum_lock_time`), UNSIGNED) AS sum_lock_time,
            CONVERT(SUM(`sum_rows_sent`), UNSIGNED) AS sum_rows_sent,
            CONVERT(SUM(`sum_rows_examined`), UNSIGNED) AS sum_rows_examined,
            CONVERT(SUM(`sum_rows_affected`), UNSIGNED) AS sum_rows_affected,
            CONVERT(SUM(`quantile_95`), UNSIGNED) AS quantile_95,
            CONVERT(SUM(`quantile_99`), UNSIGNED) AS quantile_99
        FROM
            `performance_schema`.`events_statements_summary_by_digest`
        GROUP BY
            `digest`
        ORDER BY
            SUM(`sum_timer_wait`) DESC,
            MAX(`last_seen`) DESC;
    """
    metadata_locks: str = """
        SELECT
            OBJECT_INSTANCE_BEGIN AS id,
            OBJECT_TYPE,
            OBJECT_SCHEMA AS OBJECT_SCHEMA,
            GROUP_CONCAT(OBJECT_NAME ORDER BY OBJECT_NAME) AS OBJECT_NAME,
            LOCK_TYPE,
            LOCK_STATUS,
            SOURCE AS CODE_SOURCE,
            NAME AS THREAD_SOURCE,
            PROCESSLIST_ID AS PROCESSLIST_ID,
            PROCESSLIST_USER AS PROCESSLIST_USER,
            PROCESSLIST_TIME AS PROCESSLIST_TIME,
            PROCESSLIST_INFO AS PROCESSLIST_INFO
        FROM
            `performance_schema`.`metadata_locks` mlb JOIN
            `performance_schema`.`threads` t ON mlb.OWNER_THREAD_ID = t.THREAD_ID
        WHERE
            NOT (
                OBJECT_TYPE = 'TABLE' AND
                LOCK_STATUS = 'GRANTED' AND
                LOCK_TYPE LIKE 'SHARED%' AND
                PROCESSLIST_TIME <= 2
            ) AND
            OBJECT_TYPE != 'COLUMN STATISTICS'
            $1
        GROUP BY
            THREAD_ID,
            OBJECT_TYPE,
            LOCK_TYPE,
            LOCK_STATUS
    """
    error_log: str = """
        SELECT
            logged AS timestamp,
            prio AS level,
            subsystem,
            data AS message,
            error_code
        FROM
            performance_schema.error_log
        WHERE
            data != 'Could not open log file.'
            $1
            $2
        ORDER BY
            timestamp
    """
    memory_by_user: str = """
        SELECT
            user,
            current_allocated,
            total_allocated
        FROM
            sys.memory_by_user_by_current_bytes
        WHERE
            user != "background"
    """
    memory_by_code_area: str = """
        SELECT
            SUBSTRING_INDEX( event_name, '/', 2 ) AS code_area,
            sys.format_bytes( SUM(current_alloc) ) AS current_allocated
        FROM
            sys.x$memory_global_by_current_bytes
        GROUP BY
            SUBSTRING_INDEX( event_name, '/', 2 )
        ORDER BY
            SUM( current_alloc ) DESC
    """
    memory_by_host: str = """
        SELECT
            host,
            current_allocated,
            total_allocated
        FROM
            sys.memory_by_host_by_current_bytes
        WHERE
            host != "background"
    """
    table_sizes: str = """
        SELECT
            CONCAT( t.TABLE_SCHEMA, '.', t.TABLE_NAME ) AS DATABASE_TABLE,
            t.ENGINE,
            t.ROW_FORMAT,
            s.FILE_SIZE,
            s.ALLOCATED_SIZE,
            t.DATA_LENGTH,
            t.INDEX_LENGTH,
            t.DATA_FREE,
            CONVERT(
                ROUND(( t.DATA_FREE / ( t.DATA_LENGTH + t.INDEX_LENGTH )) * 100, 2 ), UNSIGNED
            ) AS fragmentation_ratio
        FROM
            information_schema.TABLES t LEFT JOIN
            information_schema.$1 s ON CONCAT( t.TABLE_SCHEMA, '/', t.TABLE_NAME ) = s.NAME
        WHERE
            t.DATA_LENGTH + t.INDEX_LENGTH > 10 * 1024 * 1024
            AND t.TABLE_SCHEMA NOT IN ( 'mysql' )
        ORDER BY
            s.FILE_SIZE DESC,
            fragmentation_ratio DESC;
    """
    table_sizes_mariadb: str = """
        SELECT
            CONCAT( t.TABLE_SCHEMA, '.', t.TABLE_NAME ) AS DATABASE_TABLE,
            t.ENGINE,
            t.ROW_FORMAT,
            t.DATA_LENGTH,
            t.INDEX_LENGTH,
            t.DATA_FREE,
            CONVERT(
                ROUND(( t.DATA_FREE / ( t.DATA_LENGTH + t.INDEX_LENGTH )) * 100, 2 ), UNSIGNED
            ) AS fragmentation_ratio
        FROM
            information_schema.TABLES t
        WHERE
            t.DATA_LENGTH + t.INDEX_LENGTH > 10 * 1024 * 1024
            AND t.TABLE_SCHEMA NOT IN ( 'mysql' )
        ORDER BY
            ( t.DATA_LENGTH + t.INDEX_LENGTH ) DESC,
            fragmentation_ratio DESC;
    """
    databases: str = """
        SELECT
            SCHEMA_NAME
        FROM
            information_schema.SCHEMATA
        ORDER BY
            SCHEMA_NAME
    """
    innodb_metrics: str = """
        SELECT
            NAME,
            COUNT
        FROM
            information_schema.INNODB_METRICS
        WHERE
            name IN ('adaptive_hash_searches', 'adaptive_hash_searches_btree', 'trx_rseg_history_len')
    """
    active_redo_logs: str = """
        SELECT
            COUNT(*) AS count
        FROM
            performance_schema.file_instances
        WHERE
            file_name LIKE '%innodb_redo/%' AND
            file_name NOT LIKE '%_tmp'
    """
    thread_transaction_history: str = """
        SELECT
            DATE_SUB(
                NOW(),
                INTERVAL (
                    SELECT variable_value
                    FROM performance_schema.global_status
                    WHERE variable_name = 'UPTIME'
                ) - TIMER_START * 10e-13 SECOND
            ) AS start_time,
            sql_text
        FROM
            performance_schema.events_statements_history
        WHERE
            nesting_event_id = (
                SELECT EVENT_ID
                FROM performance_schema.events_transactions_current t
                WHERE t.thread_id = $1
            )
        ORDER BY
            event_id;
    """
    replication_applier_status: str = """
        SELECT
            applier_status.thread_id,
            applier_status.worker_id,
            applier_status.applying_transaction_retries_count,
            applier_status.applying_transaction_last_transient_error_timestamp,
            applier_status.applying_transaction_last_transient_error_message,
            CAST(
                (applier_status.LAST_APPLIED_TRANSACTION_END_APPLY_TIMESTAMP -
                applier_status.LAST_APPLIED_TRANSACTION_START_APPLY_TIMESTAMP) * 1000000000000
                AS CHAR
            ) AS apply_time,
            applier_status.last_applied_transaction,
            CONVERT(COALESCE(SUM(thread_events.COUNT_STAR), 0), UNSIGNED) AS total_thread_events
        FROM
            `performance_schema`.replication_applier_status_by_worker applier_status JOIN
            `performance_schema`.events_transactions_summary_by_thread_by_event_name thread_events USING (THREAD_ID)
        GROUP BY
            applier_status.thread_id
        WITH ROLLUP
        ORDER BY
            applier_status.thread_id
    """

    # Group Replication Event Horizon and Protocol
    group_replication_get_write_concurrency: str = """
        SELECT group_replication_get_write_concurrency() write_concurrency
    """
    get_group_replication_members: str = """
        SELECT
            *
        FROM
            performance_schema.replication_group_members LEFT JOIN
            performance_schema.replication_group_member_stats USING(MEMBER_ID)
    """
    get_galera_cluster_members: str = """
        SELECT
            node_uuid,
            node_name,
            node_incoming_address
        FROM
            mysql.wsrep_cluster_members
    """
    replicaset_find_replicas: str = """
        SELECT
            instance_id as id,
            address AS host,
            attributes ->> '$."replicationAccountUser"' AS user
        FROM
            mysql_innodb_cluster_metadata.instances
        WHERE
            mysql_server_uuid != @@server_uuid
    """
    determine_cluster_type_8: str = """
        SELECT
            cluster_name,
            cluster_type
        FROM
            mysql_innodb_cluster_metadata.clusters
            JOIN mysql_innodb_cluster_metadata.instances USING ( cluster_id )
        WHERE
            mysql_server_uuid = @@server_uuid
    """
    determine_cluster_type_81: str = """
        SELECT
            cluster_name,
            instance_type,
            cluster_type
        FROM
            mysql_innodb_cluster_metadata.clusters
            JOIN mysql_innodb_cluster_metadata.instances USING ( cluster_id )
            LEFT JOIN mysql_innodb_cluster_metadata.clusterset_members USING ( cluster_id )
        WHERE
            mysql_server_uuid = @@server_uuid
        ORDER BY
            view_id DESC
            LIMIT 1
    """
    get_clustersets: str = """
        SELECT
            cs.domain_name AS ClusterSet,
            GROUP_CONCAT(c.cluster_name ORDER BY c.cluster_name SEPARATOR ', ') AS Clusters
        FROM
            mysql_innodb_cluster_metadata.clustersets cs JOIN
            mysql_innodb_cluster_metadata.clusters c USING ( clusterset_id )
        GROUP BY
            cs.domain_name
    """
    get_binlog_transaction_compression_percentage: str = """
        SELECT
            compression_percentage
        FROM
            performance_schema.binary_log_transaction_compression_stats
        WHERE
            log_type = 'BINARY' AND
            compression_type = 'ZSTD'
        LIMIT
            1
    """
    status: str = """
        SHOW GLOBAL STATUS
        WHERE Variable_name IN (
            'Aborted_clients', 'Aborted_connects',
            'Binlog_cache_disk_use', 'Binlog_cache_use',
            'Com_commit', 'Com_delete', 'Com_insert', 'Com_replace',
            'Com_rollback', 'Com_select', 'Com_update',
            'Created_tmp_disk_tables', 'Created_tmp_files', 'Created_tmp_tables',
            'Innodb_buffer_pool_bytes_data', 'Innodb_buffer_pool_bytes_dirty',
            'Innodb_buffer_pool_read_requests', 'Innodb_buffer_pool_reads',
            'Innodb_buffer_pool_write_requests',
            'Innodb_checkpoint_age', 'Innodb_lsn_current', 'Innodb_os_log_written',
            'Open_tables', 'Opened_tables',
            'Queries',
            'Table_open_cache_hits', 'Table_open_cache_misses', 'Table_open_cache_overflows',
            'Threads_cached', 'Threads_connected', 'Threads_running',
            'Uptime',
            'wsrep_gcomm_uuid', 'wsrep_provider_version',
            'wsrep_cluster_size', 'wsrep_cluster_status', 'wsrep_connected',
            'wsrep_ready', 'wsrep_local_state_comment',
            'wsrep_flow_control_paused', 'wsrep_local_recv_queue_avg',
            'wsrep_local_send_queue_avg', 'wsrep_cert_deps_distance',
            'wsrep_local_commits', 'wsrep_local_cert_failures',
            'wsrep_local_bf_aborts', 'wsrep_replicated', 'wsrep_replicated_bytes',
            'wsrep_received', 'wsrep_received_bytes'
        )
    """
    variables: str = "SHOW GLOBAL VARIABLES"
    show_master_status: str = "SHOW MASTER STATUS"
    show_binary_log_status: str = "SHOW BINARY LOG STATUS"
    show_slave_status: str = "SHOW SLAVE STATUS"
    show_replica_status: str = "SHOW REPLICA STATUS"
    innodb_status: str = "SHOW ENGINE INNODB STATUS"
    show_replicas: str = "SHOW REPLICAS"
    show_slave_hosts: str = "SHOW SLAVE HOSTS"



from rich.style import Style
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Container
from textual.screen import Screen
from textual.widgets import Label, Rule, Static, TextArea
from textual.widgets.text_area import TextAreaTheme

from dolphie.Widgets.TopBar import TopBar


class ProxySQLThreadScreen(Screen):
    AUTO_FOCUS = ""

    CSS = """
        ProxySQLThreadScreen {
            background: #0a0e1b;

            & Container {
                height: auto;
            }

            & #thread_container {
                margin-top: 1;
                height: auto;
            }

            & .title {
                width: 100%;
                content-align: center middle;
                color: #bbc8e8;
                text-style: bold;
            }

            & .table {
                content-align: center middle;
                background: #101626;
                border: tall #1d253e;
                padding-left: 1;
                padding-right: 1;
                height: auto;
                width: auto;
            }

            & TextArea {
                border: tall #1d253e;
                width: 100;
                height: 35;
            }
        }
    """

    BINDINGS = [
        Binding("q", "app.pop_screen", "", show=False),
    ]

    def __init__(
        self,
        connection_status: str,
        app_version: str,
        host: str,
        thread_table: str,
        query: str,
        extended_info: str,
    ):
        super().__init__()

        self.connection_status = connection_status
        self.app_version = app_version
        self.host = host

        self.thread_table = thread_table
        self.formatted_query = query
        self.extended_info = extended_info

        dracula = TextAreaTheme.get_builtin_theme("dracula")
        dracula.base_style = Style(bgcolor="#101626")
        dracula.gutter_style = Style(color="#606e88")
        dracula.cursor_line_gutter_style = Style(color="#95a7c7", bgcolor="#20243b")
        dracula.cursor_line_style = Style(bgcolor="#20243b")
        dracula.selection_style = Style(bgcolor="#293c71")
        dracula.cursor_style = Style(bgcolor="#7a8ab2", color="#121e3a")
        dracula.syntax_styles = {
            "json.label": Style(color="#879bca", bold=True),
            "number": Style(color="#ca87a5"),
        }

        self.extended_info_text_area = TextArea(theme="dracula", show_line_numbers=True, read_only=True)
        if self.extended_info:
            self.extended_info_text_area.text = extended_info

    def on_mount(self):
        self.query_one("#thread_table").update(self.thread_table)

        if self.formatted_query:
            self.query_one("#query").update(self.formatted_query)
        else:
            self.query_one("#query_container").display = False

        if not self.extended_info:
            self.query_one("#extended_info_container").display = False

    def compose(self) -> ComposeResult:
        yield TopBar(connection_status=self.connection_status, app_version=self.app_version, host=self.host)

        with Container(id="thread_container"):
            yield Label("Thread Details", classes="title")
            yield Center(Static(id="thread_table", shrink=True, classes="table"))

        with Container(id="query_container"):
            yield Rule(line_style="heavy")
            yield Label("Query Details", classes="title")
            yield Center(Static(id="query", shrink=True, classes="table"))

        with Container(id="extended_info_container"):
            yield Rule(line_style="heavy")
            yield Label("Extended Information", classes="title")
            yield Center(self.extended_info_text_area)



from datetime import datetime, timedelta

from dolphie.Modules.Functions import format_bytes, format_number
from dolphie.Modules.MetricManager import MetricData
from dolphie.Modules.TabManager import Tab
from dolphie.Panels.Dashboard import create_system_utilization_table
from rich.style import Style
from rich.table import Table


def create_panel(tab: Tab) -> Table:
    dolphie = tab.dolphie

    global_status = dolphie.global_status
    global_variables = dolphie.global_variables
    metric_manager = dolphie.metric_manager

    ####################
    # Host Information #
    ####################
    runtime = str(datetime.now().astimezone() - dolphie.dolphie_start_time).split(".")[0]

    table_title_style = Style(color="#bbc8e8", bold=True)
    table = Table(
        show_header=False,
        box=None,
        title=f"{dolphie.panels.dashboard.formatted_key}Host Information",
        title_style=table_title_style,
    )

    table.add_column()
    table.add_column(min_width=15)
    table.add_row("[label]Version", f"{dolphie.host_distro} {dolphie.host_version}")
    table.add_row("[label]Uptime", str(timedelta(seconds=global_status["ProxySQL_Uptime"])))
    table.add_row(
        "[label]MySQL",
        (
            f"{global_variables['mysql-server_version']} "
            f"[label]Workers[/label] {global_status['MySQL_Thread_Workers']}"
        ),
    )
    if not dolphie.replay_file:
        table.add_row("[label]Runtime", runtime)

    if dolphie.worker_processing_time:
        table.add_row("[label]Latency", f"{round(dolphie.worker_processing_time, 2)}s")

    tab.dashboard_section_1.update(table)

    ######################
    # System Utilization #
    ######################
    table = create_system_utilization_table(tab)

    if table:
        tab.dashboard_section_6.update(table)

    ##########################
    # Connection Information #
    ##########################
    proxysql_connections = metric_manager.metrics.proxysql_connections

    table = Table(show_header=False, box=None, title="Connections", title_style=table_title_style)

    table.add_column()
    table.add_column(min_width=6)
    data_dict = {
        "[label]FE Connected": proxysql_connections.Client_Connections_connected.values,
        "[label]FE Non-idle": proxysql_connections.Client_Connections_non_idle.values,
        "[label]BE Connected": proxysql_connections.Server_Connections_connected.values,
        "[label]FE Created": proxysql_connections.Client_Connections_created.values,
        "[label]BE Created": proxysql_connections.Server_Connections_created.values,
    }

    fe_usage = round(
        (global_status["Client_Connections_connected"] / global_variables["mysql-max_connections"]) * 100,
        2,
    )

    metric_data = metric_manager.metrics.proxysql_multiplex_efficiency.proxysql_multiplex_efficiency_ratio
    if metric_data.values:
        if metric_data.values[-1] >= 85:
            color_code = "green"
        elif metric_data.values[-1] >= 50:
            color_code = "yellow"
        else:
            color_code = "red"

        mp_efficiency = f"[{color_code}]{metric_data.values[-1]}%[/{color_code}]"
    else:
        mp_efficiency = "N/A"

    if fe_usage >= 90:
        color_code = "red"
    elif fe_usage >= 70:
        color_code = "yellow"
    else:
        color_code = "green"

    table.add_row("[label]MP Efficiency", mp_efficiency)
    table.add_row("[label]FE Usage", f"[{color_code}]{fe_usage}%")
    table.add_row("[label]Active TRX", f"{global_status['Active_Transactions']}")
    for label, values in data_dict.items():
        value = format_number(values[-1]) if values else 0

        if "Created" in label or "Aborted" in label or "Wrong Passwd" in label:
            table.add_row(label, f"{value}/s")
        else:
            table.add_row(label, f"{value}")

    # Reuse Innodb table for connection information
    tab.dashboard_section_2.update(table)

    ####################################
    # Query Sent/Recv Rate Information #
    ####################################
    proxysql_queries_network_data = metric_manager.metrics.proxysql_queries_data_network

    table = Table(
        show_header=False,
        box=None,
        title="Query Data Rates/s",
        title_style=table_title_style,
    )

    table.add_column()
    table.add_column(min_width=9)
    data_dict = {
        "[label]FE Sent": proxysql_queries_network_data.Queries_frontends_bytes_sent.values,
        "[label]BE Sent": proxysql_queries_network_data.Queries_backends_bytes_sent.values,
        "[label]FE Recv": proxysql_queries_network_data.Queries_frontends_bytes_recv.values,
        "[label]BE Recv": proxysql_queries_network_data.Queries_backends_bytes_recv.values,
    }

    for label, values in data_dict.items():
        value = format_bytes(values[-1]) if values else 0

        if "Created" in label or "Aborted" in label or "Wrong Passwd" in label:
            table.add_row(label, f"{value}/s")
        else:
            table.add_row(label, f"{value}")

    # Reuse binary log table for connection information
    tab.dashboard_section_3.update(table)

    ###############
    # Statistics #
    ###############
    table = Table(show_header=False, box=None, title="Statistics/s", title_style=table_title_style)

    table.add_column()
    table.add_column(min_width=7)

    # Add DML statistics
    metrics = metric_manager.metrics.dml
    metric_labels = {
        "Queries": "Queries",
        "SELECT": "Com_select",
        "INSERT": "Com_insert",
        "UPDATE": "Com_update",
        "DELETE": "Com_delete",
        "REPLACE": "Com_replace",
        "COMMIT": "Com_commit",
        "ROLLBACK": "Com_rollback",
    }

    for label, metric_name in metric_labels.items():
        metric_data: MetricData = getattr(metrics, metric_name)

        if metric_data.values:
            table.add_row(f"[label]{label}", format_number(metric_data.values[-1]))
        else:
            table.add_row(f"[label]{label}", "0")

    tab.dashboard_section_4.update(table)



from dolphie.Modules.Functions import format_number
from dolphie.Modules.TabManager import Tab
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    all_columns = {
        "rule_id": {"name": "Rule ID", "format": None, "always_show": True},
        "hits": {"name": "Total Hits", "format": "number", "always_show": True},
        "hits_s": {"name": "Hits/s", "format": "number", "always_show": True},
        "apply": {"name": "Apply", "format": None, "always_show": True},
        "log": {"name": "Log", "format": None, "always_show": True},
        "flagIN": {"name": "flagIN", "format": None, "always_show": False},
        "flagOUT": {"name": "flagOUT", "format": None, "always_show": False},
        "destination_hostgroup": {
            "name": "Dest HG",
            "format": None,
            "always_show": False,
        },
        "username": {"name": "Username", "format": None, "always_show": False},
        "match_pattern": {
            "name": "Match Pattern",
            "format": None,
            "always_show": False,
        },
        "match_digest": {"name": "Match Digest", "format": None, "always_show": False},
        "schemaname": {"name": "Schema", "format": None, "always_show": False},
        "client_addr": {"name": "Client Addr", "format": None, "always_show": False},
        "proxy_addr": {"name": "Proxy Addr", "format": None, "always_show": False},
        "proxy_port": {"name": "Proxy Port", "format": None, "always_show": False},
        "digest": {"name": "Digest", "format": None, "always_show": False},
        "negate_match_pattern": {
            "name": "Negate Match",
            "format": None,
            "always_show": False,
        },
        "re_modifiers": {"name": "RE Modifiers", "format": None, "always_show": False},
        "replace_pattern": {
            "name": "Replace Pattern",
            "format": None,
            "always_show": False,
        },
        "cache_ttl": {"name": "Cache TTL", "format": None, "always_show": False},
        "cache_empty_result": {
            "name": "Cache Empty",
            "format": None,
            "always_show": False,
        },
        "cache_timeout": {
            "name": "Cache Timeout",
            "format": None,
            "always_show": False,
        },
        "reconnect": {"name": "Reconnect", "format": None, "always_show": False},
        "timeout": {"name": "Timeout", "format": None, "always_show": False},
        "retries": {"name": "Retries", "format": None, "always_show": False},
        "delay": {"name": "Delay", "format": None, "always_show": False},
        "next_query_flagIN": {
            "name": "Next flagIN",
            "format": None,
            "always_show": False,
        },
        "mirror_flagOUT": {
            "name": "Mirror flagOUT",
            "format": None,
            "always_show": False,
        },
        "mirror_hostgroup": {"name": "Mirror HG", "format": None, "always_show": False},
        "error_msg": {"name": "Error Msg", "format": None, "always_show": False},
        "OK_msg": {"name": "OK Msg", "format": None, "always_show": False},
        "sticky_conn": {"name": "Sticky Conn", "format": None, "always_show": False},
        "multiplex": {"name": "Multiplex", "format": None, "always_show": False},
        "gtid_from_hostgroup": {
            "name": "GTID from HG",
            "format": None,
            "always_show": False,
        },
        "attributes": {"name": "Attributes", "format": None, "always_show": False},
        "comment": {"name": "Comment", "format": None, "always_show": False},
    }

    # Filter only relevant columns from all_columns based on data presence or always_show flag
    columns_with_data = {
        column
        for row in dolphie.proxysql_mysql_query_rules
        for column, value in row.items()
        if value not in (None, "", "NULL", 0, "0") or all_columns[column]["always_show"]
    }

    # Build the filtered columns dictionary using only columns_with_data
    columns_filtered = {
        column: {
            "name": props["name"],
            "width": None,
            "format": props["format"],
        }
        for column, props in all_columns.items()
        if column in columns_with_data
    }

    mysql_query_rules = tab.proxysql_mysql_query_rules_datatable

    # Clear table if columns change
    if len(mysql_query_rules.columns) != len(columns_filtered):
        mysql_query_rules.clear(columns=True)

    column_keys = []
    column_names = []
    column_formats = []

    # Add columns to the datatable if it is empty
    if not mysql_query_rules.columns:
        for column_key, column_data in columns_filtered.items():
            mysql_query_rules.add_column(column_data["name"], key=column_key, width=column_data["width"])

    for column_key, column_data in columns_filtered.items():
        column_keys.append(column_key)
        column_names.append(column_data["name"])
        column_formats.append(column_data["format"])

    polling_latency = dolphie.polling_latency

    with dolphie.app.batch_update():
        # Remove stale rows first
        if dolphie.proxysql_mysql_query_rules:
            current_rows = {row["rule_id"] for row in dolphie.proxysql_mysql_query_rules}
            existing_rows = set(mysql_query_rules.rows.keys())

            rows_to_remove = existing_rows - current_rows
            if rows_to_remove:
                if len(rows_to_remove) > len(dolphie.proxysql_mysql_query_rules):
                    mysql_query_rules.clear()
                else:
                    for row_id in rows_to_remove:
                        mysql_query_rules.remove_row(row_id)
        else:
            if mysql_query_rules.row_count:
                mysql_query_rules.clear()

        for row in dolphie.proxysql_mysql_query_rules:
            row_id = row["rule_id"]

            row_values = []
            for column_key, column_format in zip(column_keys, column_formats):
                column_value = row.get(column_key)

                # Calculate the values per second for the following columns
                if column_key in ["hits_s"]:
                    previous_value = dolphie.proxysql_per_second_data.get(row_id, {}).get(column_key, 0)
                    if not previous_value:
                        column_value = "[dark_gray]0"
                    else:
                        current_value = int(column_value or 0)
                        value_diff = current_value - previous_value
                        column_value = round(value_diff / polling_latency) if polling_latency > 0 else 0

                    dolphie.proxysql_per_second_data.setdefault(row_id, {})[column_key] = int(row.get(column_key, 0))

                if column_key in ["apply", "log"]:
                    column_value = "Yes" if column_value == "1" else "No"

                if column_format == "number":
                    column_value = format_number(column_value)

                if column_value is None:
                    column_value = "[dark_gray]N/A"
                elif column_value == "0":
                    column_value = "[dark_gray]0"

                row_values.append(column_value)

            if row_id in mysql_query_rules.rows:
                datatable_row = mysql_query_rules.get_row(row_id)

                for column_id, column_key in enumerate(column_keys):
                    if row_values[column_id] != datatable_row[column_id]:
                        mysql_query_rules.update_cell(row_id, column_key, row_values[column_id])
            else:
                mysql_query_rules.add_row(*row_values, key=row_id)

    tab.proxysql_mysql_query_rules_title.update(
        f"{dolphie.panels.proxysql_mysql_query_rules.title} "
        f"([$highlight]{mysql_query_rules.row_count}[/$highlight])"
    )



from dolphie.Modules.Functions import format_number
from dolphie.Modules.TabManager import Tab
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    columns = {
        "Command": {"name": "Command", "width": None, "format": None},
        "Total_cnt": {"name": "Total", "width": 10, "format": "number"},
        "Total_cnt_s": {"name": "Total/s", "width": 10, "format": "number"},
        "cnt_100us": {"name": "100μs/s", "width": 8, "format": "number"},
        "cnt_500us": {"name": "500μs/s", "width": 8, "format": "number"},
        "cnt_1ms": {"name": "1ms/s", "width": 8, "format": "number"},
        "cnt_5ms": {"name": "5ms/s", "width": 8, "format": "number"},
        "cnt_10ms": {"name": "10ms/s", "width": 8, "format": "number"},
        "cnt_50ms": {"name": "50ms/s", "width": 8, "format": "number"},
        "cnt_100ms": {"name": "100ms/s", "width": 8, "format": "number"},
        "cnt_500ms": {"name": "500ms/s", "width": 8, "format": "number"},
        "cnt_1s": {"name": "1s/s", "width": 8, "format": "number"},
        "cnt_5s": {"name": "5s/s", "width": 8, "format": "number"},
        "cnt_10s": {"name": "10s/s", "width": 8, "format": "number"},
        "cnt_INFs": {"name": "10s+/s", "width": 8, "format": "number"},
    }

    command_stats = tab.proxysql_command_stats_datatable

    column_keys = []
    column_names = []
    column_formats = []

    # Add columns to the datatable if it is empty
    if not command_stats.columns:
        for column_key, column_data in columns.items():
            command_stats.add_column(column_data["name"], key=column_key, width=column_data["width"])

    for column_key, column_data in columns.items():
        column_keys.append(column_key)
        column_names.append(column_data["name"])
        column_formats.append(column_data["format"])

    polling_latency = dolphie.polling_latency

    with dolphie.app.batch_update():
        # Remove stale rows first
        if dolphie.proxysql_command_stats:
            current_rows = {row["Command"] for row in dolphie.proxysql_command_stats}
            existing_rows = set(command_stats.rows.keys())

            rows_to_remove = existing_rows - current_rows
            if rows_to_remove:
                if len(rows_to_remove) > len(dolphie.proxysql_command_stats):
                    command_stats.clear()
                else:
                    for row_id in rows_to_remove:
                        command_stats.remove_row(row_id)
        else:
            if command_stats.row_count:
                command_stats.clear()

        for row in dolphie.proxysql_command_stats:
            row_id = row["Command"]

            row_values = []
            for column_key, column_format in zip(column_keys, column_formats):
                column_value = row.get(column_key)

                # Calculate the values per second for the following columns
                if "cnt_" in column_key:
                    current_value = int(column_value or 0)
                    previous_value = dolphie.proxysql_per_second_data.get(row_id, {}).get(column_key, 0)
                    if not previous_value:
                        column_value = 0
                    else:
                        value_diff = current_value - previous_value
                        column_value = round(value_diff / polling_latency) if polling_latency > 0 else 0

                    dolphie.proxysql_per_second_data.setdefault(row_id, {})[column_key] = current_value

                if column_format == "number":
                    column_value = format_number(column_value)

                if column_value == "0":
                    column_value = "[dark_gray]0"

                row_values.append(column_value)

            if row_id in command_stats.rows:
                datatable_row = command_stats.get_row(row_id)

                for column_id, column_key in enumerate(column_keys):
                    if row_values[column_id] != datatable_row[column_id]:
                        command_stats.update_cell(row_id, column_key, row_values[column_id])
            else:
                command_stats.add_row(*row_values, key=row_id)

    tab.proxysql_command_stats_title.update(
        f"{dolphie.panels.proxysql_command_stats.title} " f"([$highlight]{command_stats.row_count}[/$highlight])"
    )



import os
import re
from decimal import Decimal

import charset_normalizer
from pygments.style import Style
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Name,
    Number,
    Operator,
    Punctuation,
    String,
    Whitespace,
)
from rich.markup import escape as markup_escape
from rich.syntax import Syntax


class NordModifiedTheme(Style):
    nord0 = "#0f1525"
    nord1 = "#3b4252"
    nord2 = "#434c5e"
    nord3 = "#4c566a"
    nord3_bright = "#616e87"

    nord4 = "#d8dee9"
    nord5 = "#e5e9f0"
    nord6 = "#eceff4"

    nord7 = "#8fbcbb"
    nord8 = "#88c0d0"
    nord9 = "#879bca"
    nord10 = "#5e81ac"

    nord11 = "#bf616a"
    nord12 = "#d08770"
    nord13 = "#81c194"
    nord14 = "#ac8bdd"
    nord15 = "#ca87a5"

    background_color = nord0
    default = nord4

    styles = {
        Whitespace: nord4,
        Comment: f"italic {nord3_bright}",
        Comment.Preproc: nord10,
        Keyword: f"bold {nord9}",
        Keyword.Pseudo: f"nobold {nord9}",
        Keyword.Type: f"nobold {nord9}",
        Operator: nord9,
        Operator.Word: f"bold {nord9}",
        Name: nord4,
        Name.Builtin: nord9,
        Name.Function: nord8,
        Name.Class: nord7,
        Name.Namespace: nord7,
        Name.Exception: nord11,
        Name.Variable: nord4,
        Name.Constant: nord7,
        Name.Label: nord7,
        Name.Entity: nord12,
        Name.Attribute: nord7,
        Name.Tag: nord9,
        Name.Decorator: nord12,
        Punctuation: nord6,
        String: nord14,
        String.Doc: nord3_bright,
        String.Interpol: nord14,
        String.Escape: nord13,
        String.Regex: nord13,
        String.Symbol: nord14,
        String.Other: nord14,
        Number: nord15,
        Generic.Heading: f"bold {nord8}",
        Generic.Subheading: f"bold {nord8}",
        Generic.Deleted: nord11,
        Generic.Inserted: nord14,
        Generic.Error: nord11,
        Generic.Emph: "italic",
        Generic.Strong: "bold",
        Generic.Prompt: f"bold {nord3}",
        Generic.Output: nord4,
        Generic.Traceback: nord11,
        Error: nord11,
    }


def format_query(query: str, minify: bool = True) -> Syntax:
    if not query:
        return Syntax(code="", lexer="sql", word_wrap=True, theme=NordModifiedTheme)

    query = markup_escape(re.sub(r"\s+", " ", query)) if minify else query

    formatted_query = Syntax(code=query, lexer="sql", word_wrap=True, theme=NordModifiedTheme)

    return formatted_query


def minify_query(query: str) -> str:
    if not query:
        return ""

    return markup_escape(re.sub(r"\s+", " ", query))


def format_bytes(bytes_value, color=True, decimal=2):
    if isinstance(bytes_value, str):
        return bytes_value

    units = ["B", "KB", "MB", "GB", "TB"]
    unit_index = 0

    while bytes_value >= 1024 and unit_index < len(units) - 1:
        bytes_value /= 1024
        unit_index += 1

    # Format with specified rounding precision
    formatted_value = f"{bytes_value:.{decimal}f}"

    # Remove unnecessary ".00" if rounding results in a whole number
    if formatted_value.endswith(f".{'0' * decimal}"):
        formatted_value = formatted_value[: -decimal - 1]

    if bytes_value == 0:
        return "0"
    elif color:
        return f"{formatted_value}[highlight]{units[unit_index]}[/highlight]"
    else:
        return f"{formatted_value}{units[unit_index]}"


def format_time(time: int, picoseconds=False):
    if time is None:
        return "N/A"

    seconds = time / 1e12 if picoseconds else time

    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    return f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}"


def format_picoseconds(ps: int) -> str:
    units = [
        ("d", 24 * 60 * 60 * 1_000_000_000_000),  # days
        ("h", 60 * 60 * 1_000_000_000_000),  # hours
        ("m", 60 * 1_000_000_000_000),  # minutes
        ("s", 1_000_000_000_000),  # seconds
        ("ms", 1_000_000_000),  # milliseconds
        ("µs", 1_000_000),  # microseconds
        ("ns", 1_000),  # nanoseconds
        ("ps", 1),  # picoseconds
    ]

    for unit, factor in units:
        if ps >= factor:
            value = ps / factor
            return f"{value:.2f}[highlight]{unit}"
    return "[dark_gray]0"


def load_host_cache_file(host_cache_file: str):
    host_cache = {}
    if os.path.exists(host_cache_file):
        with open(host_cache_file) as file:
            for line in file:
                line = line.strip()
                error_message = f"Host cache entry '{line}' is not properly formatted! Format: ip=hostname"

                if "=" not in line:
                    raise Exception(error_message)

                host, hostname = line.split("=", maxsplit=1)
                host = host.strip()
                hostname = hostname.strip()

                if not host or not hostname:
                    raise Exception(error_message)

                host_cache[host] = hostname

    return host_cache


def detect_encoding(text):
    # Since BLOB/BINARY data can be involved, we need to auto-detect what the encoding is
    # for queries since it can be anything. If I let pymsql use unicode by default I got
    # consistent crashes due to unicode errors for utf8 so we have to go this route
    result = charset_normalizer.detect(text)
    encoding = result["encoding"]

    if encoding is None:
        encoding = "latin1"
    elif encoding == "utf-16be":
        encoding = "utf-8"

    return encoding


def round_num(n, decimal=2):
    n = Decimal(n)
    return n.to_integral() if n == n.to_integral() else round(n.normalize(), decimal)


# This is from https://pypi.org/project/numerize
def format_number(n, decimal=2, color=True):
    if not n or n == "0":
        return "0"

    # fmt: off
    sufixes = ["", "K", "M", "B", "T", "Qa", "Qu", "S", "Oc", "No",
               "D", "Ud", "Dd", "Td", "Qt", "Qi", "Se", "Od", "Nd", "V",
               "Uv", "Dv", "Tv", "Qv", "Qx", "Sx", "Ox", "Nx", "Tn", "Qa",
               "Qu", "S", "Oc", "No", "D", "Ud", "Dd", "Td", "Qt", "Qi",
               "Se", "Od", "Nd", "V", "Uv", "Dv", "Tv", "Qv", "Qx", "Sx",
               "Ox", "Nx", "Tn", "x", "xx", "xxx", "X", "XX", "XXX", "END"]

    sci_expr = [1e0, 1e3, 1e6, 1e9, 1e12, 1e15, 1e18, 1e21, 1e24, 1e27,
                1e30, 1e33, 1e36, 1e39, 1e42, 1e45, 1e48, 1e51, 1e54, 1e57,
                1e60, 1e63, 1e66, 1e69, 1e72, 1e75, 1e78, 1e81, 1e84, 1e87,
                1e90, 1e93, 1e96, 1e99, 1e102, 1e105, 1e108, 1e111, 1e114, 1e117,
                1e120, 1e123, 1e126, 1e129, 1e132, 1e135, 1e138, 1e141, 1e144, 1e147,
                1e150, 1e153, 1e156, 1e159, 1e162, 1e165, 1e168, 1e171, 1e174, 1e177]
    # fmt: on

    # Convert string to a number format if needed
    if isinstance(n, str):
        try:
            n = float(n)
        except ValueError:
            return n

    n = abs(n)
    for x in range(len(sci_expr)):
        if n >= sci_expr[x] and n < sci_expr[x + 1]:
            sufix = sufixes[x]
            num = str(round_num(n / sci_expr[x], decimal)) if n >= 1000.0 else str(round_num(n, 0))
            if color:
                return f"{num}[highlight]{sufix}[/highlight]" if sufix else num
            else:
                return f"{num}{sufix}" if sufix else num


def format_sys_table_memory(data):
    parsed_data = data.strip().split(" ")
    if len(parsed_data) == 2:
        value, suffix = parsed_data[0], parsed_data[1][:1]

        if value == "0":
            suffix = ""
        elif suffix != "b":
            suffix += "B"
        elif suffix == "b":
            suffix = "B"

        return f"{value}[highlight]{suffix}"

    return data


def escape_markup(text: str) -> str:
    """Escape Rich markup characters in the given text.

    Args:
        text (str): The text to escape.

    Returns:
        str: The escaped text.
    """
    return text.replace("[", r"\[")



from loguru import logger
from rich import box
from rich.table import Table

from dolphie.Modules.Functions import format_query, minify_query


class ManualException(Exception):
    def __init__(self, reason: str, query: str = "", code: int = None):
        self.reason = reason
        self.query = query
        self.code = code

    def output(self):
        table_exception = Table(box=box.SQUARE, show_header=True, style="#ec8888")

        table_exception.add_column("MySQL Connection Error", overflow="fold")

        logger_message = []

        if self.query:
            table_exception.add_row("[red]Failed to execute query:[/red]")
            table_exception.add_row(format_query(self.query, minify=False))
            table_exception.add_row("")

            logger_message.append(f"Query: {minify_query(self.query)}")

        if self.reason:
            # pymysql for some reason returns "ny connections" instead of "Too many connections"
            if isinstance(self.reason, str):
                self.reason = self.reason.replace("ny connections", "Too many connections")

            table_exception.add_row(self.reason)
            logger_message.append(self.reason)

        if logger_message:
            logger.critical("\n".join(logger_message))

        return table_exception



from dolphie.DataTypes import ProcesslistThread, ProxySQLProcesslistThread
from dolphie.Modules.Functions import format_query
from dolphie.Modules.Queries import ProxySQLQueries
from dolphie.Modules.TabManager import Tab
from rich.syntax import Syntax
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    columns = [
        {"name": "Thread ID", "field": "id", "width": 11},
        {"name": "Hostgroup", "field": "hostgroup", "width": 9},
        {"name": "Username", "field": "user", "width": 20},
    ]

    if dolphie.show_additional_query_columns:
        columns.extend(
            [
                {"name": "Frontend Host", "field": "frontend_host", "width": 25},
            ]
        )

    columns.extend(
        [
            {"name": "Backend Host", "field": "host", "width": 25},
            {"name": "Database", "field": "db", "width": 17},
            {"name": "Command", "field": "command", "width": 8},
            {"name": "Age", "field": "formatted_time", "width": 9},
            {"name": "Query", "field": "formatted_query", "width": None},
            {"name": "time_seconds", "field": "time", "width": 0},
        ]
    )

    # Refresh optimization
    query_length_max = 300
    processlist_datatable = tab.processlist_datatable

    # Clear table if columns change
    if len(processlist_datatable.columns) != len(columns):
        processlist_datatable.clear(columns=True)

    column_names = []
    column_fields = []

    # Add columns to the datatable if it is empty
    if not processlist_datatable.columns:
        for column_data in columns:
            processlist_datatable.add_column(column_data["name"], key=column_data["name"], width=column_data["width"])

    for column_data in columns:
        column_names.append(column_data["name"])
        column_fields.append(column_data["field"])

    threads_to_render: dict[str, ProxySQLProcesslistThread] = {}
    if dolphie.replay_file:
        for thread_id, thread in dolphie.processlist_threads.items():
            thread: ProxySQLProcesslistThread

            # Check each filter condition and skip thread if it doesn't match
            if dolphie.user_filter and dolphie.user_filter != thread.user:
                continue

            if dolphie.db_filter and dolphie.db_filter != thread.db:
                continue

            if dolphie.host_filter and dolphie.host_filter not in thread.host:
                continue

            if dolphie.query_time_filter and thread.time < dolphie.query_time_filter:
                continue

            if dolphie.query_filter and dolphie.query_filter not in thread.formatted_query.code:
                continue

            if dolphie.hostgroup_filter and dolphie.hostgroup_filter != thread.hostgroup:
                continue

            # If all checks passed, add it to the visible list
            threads_to_render[thread_id] = thread
    else:
        # Not a replay file, so fetch_data() already filtered.
        threads_to_render = dolphie.processlist_threads

    changed = False

    with dolphie.app.batch_update():
        # Remove stale rows first
        if threads_to_render:
            rows_to_remove = set(processlist_datatable.rows.keys()) - set(threads_to_render.keys())
            if rows_to_remove:
                changed = True
                if len(rows_to_remove) > len(threads_to_render):
                    processlist_datatable.clear()
                else:
                    for id in rows_to_remove:
                        processlist_datatable.remove_row(id)
        else:
            if processlist_datatable.row_count:
                changed = True
                processlist_datatable.clear()

        for thread_id, thread in threads_to_render.items():
            thread: ProxySQLProcesslistThread

            row_values = []
            for column_field in column_fields:
                value = getattr(thread, column_field)
                if column_field == "formatted_query" and isinstance(value, Syntax):
                    value = format_query(value.code[:query_length_max])
                row_values.append(value)

            if thread_id in processlist_datatable.rows:
                datatable_row = processlist_datatable.get_row(thread_id)

                for column_id, (column_name, column_field) in enumerate(zip(column_names, column_fields)):
                    new_val = row_values[column_id]
                    old_val = datatable_row[column_id]

                    # Compare text content for Syntax objects
                    cmp_new = new_val.code if isinstance(new_val, Syntax) else new_val
                    cmp_old = old_val.code if isinstance(old_val, Syntax) else old_val

                    if cmp_new != cmp_old or column_field == "formatted_time" or column_field == "time":
                        changed = True
                        processlist_datatable.update_cell(
                            thread_id,
                            column_name,
                            new_val,
                            update_width=(column_field == "formatted_query"),
                        )
            else:
                changed = True
                processlist_datatable.add_row(*row_values, key=thread_id)

        if changed:
            processlist_datatable.sort("time_seconds", reverse=dolphie.sort_by_time_descending)

    if dolphie.replay_file:
        dolphie.processlist_threads = threads_to_render

    tab.processlist_title.update(
        f"{dolphie.panels.processlist.title} " f"([$highlight]{processlist_datatable.row_count}[/$highlight])"
    )


def fetch_data(tab: Tab) -> dict[str, ProcesslistThread]:
    dolphie = tab.dolphie

    ########################
    # WHERE clause filters #
    ########################
    where_clause = []

    # Filter out idle threads if specified
    if not dolphie.show_idle_threads:
        where_clause.append("command NOT IN ('Sleep', 'Connecting client')")

    # Filter user
    if dolphie.user_filter:
        where_clause.append(f"user = '{dolphie.user_filter}'")

    # Filter database
    if dolphie.db_filter:
        where_clause.append(f"db = '{dolphie.db_filter}'")

    # Filter hostname/IP
    if dolphie.host_filter:
        where_clause.append(f"srv_host = '{dolphie.host_filter}'")

    # Filter time
    if dolphie.query_time_filter:
        # Convert to milliseconds
        where_clause.append(f"time_ms >= '{dolphie.query_time_filter * 1000}'")

    # Filter query
    if dolphie.query_filter:
        where_clause.append(f"info LIKE '%%{dolphie.query_filter}%%'")

    # Filter hostgroup
    if dolphie.hostgroup_filter:
        where_clause.append(f"hostgroup = '{dolphie.hostgroup_filter}'")

    # Add in our dynamic WHERE clause for filtering
    if where_clause:
        processlist_query = ProxySQLQueries.processlist.replace("$1", " AND ".join(where_clause))
    else:
        processlist_query = ProxySQLQueries.processlist.replace("$1", "1=1")

    processlist_threads = {}
    # Run the processlist query
    dolphie.main_db_connection.execute(processlist_query)
    threads = dolphie.main_db_connection.fetchall()

    for thread in threads:
        # Don't include Dolphie's threads
        if (
            dolphie.main_db_connection.connection_id == thread["id"]
            or dolphie.secondary_db_connection.connection_id == thread["id"]
        ):
            continue

        thread["frontend_host"] = dolphie.get_hostname(thread["frontend_host"])
        thread["backend_host"] = dolphie.get_hostname(thread["backend_host"])
        thread["query"] = thread["query"] or ""

        processlist_threads[str(thread["id"])] = ProxySQLProcesslistThread(thread)

    return processlist_threads



from __future__ import annotations

from dolphie.Modules.Functions import format_query, format_time
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Modules.TabManager import Tab
from rich.syntax import Syntax
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    columns = {
        "OBJECT_TYPE": {"name": "Object Type", "width": 13},
        "OBJECT_SCHEMA": {"name": "Object Schema", "width": 13},
        "OBJECT_NAME": {"name": "Object Name", "width": 25},
        "LOCK_TYPE": {"name": "Lock Type", "width": 20},
        "LOCK_STATUS": {"name": "Lock Status", "width": 11},
        "CODE_SOURCE": {"name": "Code Source", "width": 15},
        "THREAD_SOURCE": {"name": "Thread Source", "width": 15},
        "PROCESSLIST_ID": {"name": "Process ID", "width": 13},
        "PROCESSLIST_USER": {"name": "User", "width": 20},
        "PROCESSLIST_TIME": {"name": "Age", "width": 8},
        "PROCESSLIST_INFO": {"name": "Query", "width": None},
    }

    # Refresh optimization
    query_length_max = 300
    metadata_locks_datatable = tab.metadata_locks_datatable

    column_keys = []
    column_names = []
    column_widths = []

    if not metadata_locks_datatable.columns:
        for column_key, column_data in columns.items():
            metadata_locks_datatable.add_column(
                column_data["name"], key=column_data["name"], width=column_data["width"]
            )

    for column_key, column_data in columns.items():
        column_keys.append(column_key)
        column_names.append(column_data["name"])
        column_widths.append(column_data["width"])

    changed = False

    with dolphie.app.batch_update():
        # Remove stale rows first
        if dolphie.metadata_locks:
            rows_to_remove = set(metadata_locks_datatable.rows.keys()) - {
                str(lock["id"]) for lock in dolphie.metadata_locks
            }
            if rows_to_remove:
                changed = True
                if len(rows_to_remove) > len(dolphie.metadata_locks):
                    metadata_locks_datatable.clear()
                else:
                    for id in rows_to_remove:
                        metadata_locks_datatable.remove_row(id)
        else:
            if metadata_locks_datatable.row_count:
                changed = True
                metadata_locks_datatable.clear()

        for lock in dolphie.metadata_locks:
            lock_id = str(lock["id"])
            row_height = 1

            row_values = []
            for column_key, column_name, column_width in zip(column_keys, column_names, column_widths):
                column_value = lock[column_key]

                # Get height of row based on the how many objects are in the OBJECT_NAME field
                if (
                    column_key == "OBJECT_NAME"
                    and column_value
                    and len(column_value) > column_width
                    and "," in column_value
                ):
                    object_names = [object_name[:column_width] for object_name in column_value.split(",")]
                    value = "\n".join(object_names)
                    row_height = len(object_names)
                else:
                    value = format_value(lock, column_key, column_value)

                # Truncate query Syntax objects
                if column_key == "PROCESSLIST_INFO" and isinstance(value, Syntax):
                    value = format_query(value.code[:query_length_max])

                row_values.append(value)

            if lock_id in metadata_locks_datatable.rows:
                datatable_row = metadata_locks_datatable.get_row(lock_id)

                for column_id, column_name in enumerate(column_names):
                    new_val = row_values[column_id]
                    old_val = datatable_row[column_id]

                    # Compare text content for Syntax objects
                    cmp_new = new_val.code if isinstance(new_val, Syntax) else new_val
                    cmp_old = old_val.code if isinstance(old_val, Syntax) else old_val

                    if cmp_new != cmp_old:
                        changed = True
                        metadata_locks_datatable.update_cell(lock_id, column_name, new_val)
            else:
                changed = True
                metadata_locks_datatable.add_row(*row_values, key=lock_id, height=row_height)

        if changed:
            metadata_locks_datatable.sort("Age", reverse=dolphie.sort_by_time_descending)

    tab.metadata_locks_title.update(
        f"{dolphie.panels.metadata_locks.title} " f"([$highlight]{metadata_locks_datatable.row_count}[/$highlight])"
    )


def fetch_data(tab: Tab) -> list[dict[str, int | str]]:
    dolphie = tab.dolphie

    ########################
    # WHERE clause filters #
    ########################
    where_clause = []

    # Filter user
    if dolphie.user_filter:
        where_clause.append(f"processlist_user = '{dolphie.user_filter}'")

    # Filter database
    if dolphie.db_filter:
        where_clause.append(f"processlist_db = '{dolphie.db_filter}'")

    # Filter hostname/IP
    if dolphie.host_filter:
        # Have to use LIKE since there's a port at the end
        where_clause.append(f"processlist_host LIKE '{dolphie.host_filter}%'")

    # Filter time
    if dolphie.query_time_filter:
        where_clause.append(f"processlist_time >= '{dolphie.query_time_filter}'")

    # Filter query
    if dolphie.query_filter:
        where_clause.append(f"(processlist_info LIKE '%{dolphie.query_filter}%')")

    if where_clause:
        # Add in our dynamic WHERE clause for filtering
        query = MySQLQueries.metadata_locks.replace("$1", "AND " + " AND ".join(where_clause))
    else:
        query = MySQLQueries.metadata_locks.replace("$1", "")

    dolphie.main_db_connection.execute(query)
    threads = dolphie.main_db_connection.fetchall()

    return threads


def format_value(lock: dict, column_key: str, value: str) -> str:
    formatted_value = value

    # OBJECT_NAME is in the format "schema/table" sometimes where OBJECT_SCHEMA is empty,
    # so I want to split OBJECT_NAME and correct it if necessary
    if column_key == "OBJECT_SCHEMA" and not value and lock["OBJECT_NAME"] and "/" in lock["OBJECT_NAME"]:
        formatted_value = lock["OBJECT_NAME"].split("/")[0]
    elif column_key == "OBJECT_NAME" and value and "/" in value:
        formatted_value = value.split("/")[1]
    elif value is None or value == "":
        formatted_value = "[dark_gray]N/A"
    elif column_key == "PROCESSLIST_INFO":
        formatted_value = format_query(value)
    elif column_key == "LOCK_STATUS":
        if value == "GRANTED":
            formatted_value = f"[green]{value}[/green]"
        elif value == "PENDING":
            formatted_value = f"[red]{value}[/red]"
    elif column_key == "LOCK_TYPE":
        if value == "EXCLUSIVE":
            formatted_value = f"[yellow]{value}[/yellow]"
    elif column_key == "PROCESSLIST_TIME":
        formatted_value = format_time(value)
    elif column_key == "CODE_SOURCE":
        formatted_value = value.split(":")[0]
    elif column_key == "THREAD_SOURCE":
        formatted_value = value.split("/")[-1]

        if formatted_value == "one_connection":
            formatted_value = "user_connection"

    return formatted_value



import os
from datetime import datetime

from dolphie.Modules.Functions import format_bytes, format_number, format_time
from dolphie.Modules.TabManager import Tab
from textual.widgets import DataTable


def create_panel(tab: Tab):
    dolphie = tab.dolphie

    update_file_io_by_instance(tab)
    update_table_io_waits_summary_by_table(tab)

    if dolphie.replay_file:
        time = dolphie.global_status.get("replay_pfs_metrics_last_reset_time", 0)
    else:
        time = (
            (datetime.now().astimezone() - dolphie.pfs_metrics_last_reset_time).total_seconds()
            if dolphie.pfs_metrics_last_reset_time
            else 0
        )
    tab.pfs_metrics_delta.label = f"Delta since last reset ([$light_blue]{format_time(time)}[/$light_blue])"


def update_table_io_waits_summary_by_table(tab: Tab) -> DataTable:
    dolphie = tab.dolphie
    datatable = tab.pfs_metrics_table_io_waits_datatable

    if not dolphie.table_io_waits_data or not dolphie.table_io_waits_data.filtered_data:
        datatable.display = False
        tab.pfs_metrics_tabs.get_tab("pfs_metrics_table_io_waits_tab").label = (
            "Table I/O Waits ([$highlight]0[/$highlight])"
        )

        return

    datatable.display = True

    columns = {
        "Table": {"field": "TABLE_NAME", "width": None},
        "Total": {"field": ["COUNT_STAR", "SUM_TIMER_WAIT"], "width": 23},
        "Fetch": {"field": ["COUNT_FETCH", "SUM_TIMER_FETCH"], "width": 23},
        "Insert": {"field": ["COUNT_INSERT", "SUM_TIMER_INSERT"], "width": 23},
        "Update": {"field": ["COUNT_UPDATE", "SUM_TIMER_UPDATE"], "width": 23},
        "Delete": {"field": ["COUNT_DELETE", "SUM_TIMER_DELETE"], "width": 23},
        "wait_time_ps": {"field": "SUM_TIMER_WAIT", "width": 0},
    }

    column_names = []
    column_fields = []

    # Add columns to the datatable if it is empty
    if not datatable.columns:
        for column_key, column_data in columns.items():
            column_width = column_data["width"]
            datatable.add_column(column_key, key=column_key, width=column_width)

    for column_key, column_data in columns.items():
        column_names.append(column_key)
        column_fields.append(column_data["field"])

    data = dolphie.table_io_waits_data.filtered_data
    use_total = tab.pfs_metrics_radio_set.pressed_button.id == "pfs_metrics_total"

    changed = False

    with dolphie.app.batch_update():
        # Remove stale rows first
        if data:
            current_rows = set(data.keys())
            existing_rows = set(datatable.rows.keys())

            rows_to_remove = existing_rows - current_rows
            if rows_to_remove:
                changed = True
                if len(rows_to_remove) > len(data):
                    datatable.clear()
                else:
                    for row_id in rows_to_remove:
                        datatable.remove_row(row_id)
        else:
            if datatable.row_count:
                changed = True
                datatable.clear()

        mode = "t" if use_total else "d"

        for file_name, metrics in data.items():
            row_id = file_name

            row_values = [file_name]
            for column_name, field in zip(column_names, column_fields):
                if column_name == "Table":
                    continue

                if isinstance(field, list):
                    count_field, wait_time_field = field
                    count_value = metrics.get(count_field, {}).get(mode, 0)
                    wait_time_value = metrics.get(wait_time_field, {}).get(mode, 0)

                    if count_value and wait_time_value:
                        formatted_time = format_time(wait_time_value, picoseconds=True)
                        column_value = f"{formatted_time} ({format_number(count_value)})"
                    else:
                        column_value = "[dark_gray]N/A"
                else:
                    column_value = metrics.get(field, {}).get(mode, 0)

                row_values.append(column_value)

            if row_id in datatable.rows:
                datatable_row = datatable.get_row(row_id)

                for column_id, column_name in enumerate(column_names):
                    if row_values[column_id] != datatable_row[column_id]:
                        changed = True
                        datatable.update_cell(row_id, column_name, row_values[column_id])
            else:
                changed = True
                datatable.add_row(*row_values, key=row_id)

        if changed:
            datatable.sort("wait_time_ps", reverse=True)

    # Update the title to reflect the number of active rows
    tab.pfs_metrics_tabs.get_tab("pfs_metrics_table_io_waits_tab").label = (
        f"Table I/O Waits ([$highlight]{datatable.row_count}[/$highlight])"
    )


def update_file_io_by_instance(tab: Tab) -> DataTable:
    dolphie = tab.dolphie
    datatable = tab.pfs_metrics_file_io_datatable

    if not dolphie.file_io_data or not dolphie.file_io_data.filtered_data:
        datatable.display = False
        tab.pfs_metrics_tabs.get_tab("pfs_metrics_file_io_tab").label = "File I/O ([$highlight]0[/$highlight])"

        return

    datatable.display = True

    columns = {
        "File or Table": {"field": "FILE_NAME", "width": None},
        "Wait Time": {"field": "SUM_TIMER_WAIT", "width": 10, "format": "time"},
        "Read Ops": {"field": "COUNT_READ", "width": 10, "format": "number"},
        "Write Ops": {"field": "COUNT_WRITE", "width": 10, "format": "number"},
        "Misc Ops": {"field": "COUNT_MISC", "width": 10, "format": "number"},
        "Read Bytes": {
            "field": "SUM_NUMBER_OF_BYTES_READ",
            "width": 10,
            "format": "bytes",
        },
        "Write Bytes": {
            "field": "SUM_NUMBER_OF_BYTES_WRITE",
            "width": 11,
            "format": "bytes",
        },
        "wait_time_ps": {"field": "SUM_TIMER_WAIT", "width": 0},
    }

    # Pre-extract column information for efficient row processing
    column_names = []
    column_fields = []
    column_formats = []

    # Add columns to the datatable if it is empty
    if not datatable.columns:
        for column_key, column_data in columns.items():
            column_width = column_data["width"]
            datatable.add_column(column_key, key=column_key, width=column_width)

    for column_key, column_data in columns.items():
        column_names.append(column_key)
        column_fields.append(column_data["field"])
        column_formats.append(column_data.get("format"))

    data = dolphie.file_io_data.filtered_data
    use_total = tab.pfs_metrics_radio_set.pressed_button.id == "pfs_metrics_total"

    changed = False

    with dolphie.app.batch_update():
        # Remove stale rows first
        if data:
            current_rows = set(data.keys())
            existing_rows = set(datatable.rows.keys())

            rows_to_remove = existing_rows - current_rows
            if rows_to_remove:
                changed = True
                if len(rows_to_remove) > len(data):
                    datatable.clear()
                else:
                    for row_id in rows_to_remove:
                        datatable.remove_row(row_id)
        else:
            if datatable.row_count:
                changed = True
                datatable.clear()

        mode = "t" if use_total else "d"

        for file_name, metrics in data.items():
            row_id = file_name

            table_match = dolphie.file_io_data.table_pattern.search(file_name)
            if file_name.endswith("/mysql.ibd"):
                file_name = f"[dark_gray]{os.path.dirname(file_name)}[/dark_gray]/{os.path.basename(file_name)}"
            elif table_match:
                file_name = f"{table_match.group(1)}.{table_match.group(2)}"
            elif "/" in file_name:
                file_name = f"[dark_gray]{os.path.dirname(file_name)}[/dark_gray]/{os.path.basename(file_name)}"
            else:
                file_name = f"[b][light_blue][[/light_blue][/b][highlight]{file_name}[b][light_blue]][/light_blue][/b]"

            row_values = [file_name]
            for column_name, field, column_format in zip(column_names, column_fields, column_formats):
                if field == "FILE_NAME":
                    continue

                column_value = metrics.get(field, {}).get(mode, 0)

                if column_format == "time":
                    column_value = format_time(column_value, picoseconds=True)
                elif column_value == 0 or column_value is None:
                    column_value = 0 if column_name == "wait_time_ps" else "[dark_gray]0"
                elif column_format == "number":
                    column_value = format_number(column_value)
                elif column_format == "bytes":
                    column_value = format_bytes(column_value)

                row_values.append(column_value)

            if row_id in datatable.rows:
                datatable_row = datatable.get_row(row_id)

                for column_id, column_name in enumerate(column_names):
                    if row_values[column_id] != datatable_row[column_id]:
                        changed = True
                        datatable.update_cell(row_id, column_name, row_values[column_id])
            else:
                changed = True
                datatable.add_row(*row_values, key=row_id)

        if changed:
            datatable.sort("wait_time_ps", reverse=True)

    # Update the title to reflect the number of active rows
    tab.pfs_metrics_tabs.get_tab("pfs_metrics_file_io_tab").label = (
        f"File I/O ([$highlight]{datatable.row_count}[/$highlight])"
    )



from functools import partial
from typing import TYPE_CHECKING

from textual.command import DiscoveryHit, Hit, Provider

if TYPE_CHECKING:
    from dolphie.App import DolphieApp


class CommandPaletteCommands(Provider):
    """Command palette commands based on connection source."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dolphie_app: DolphieApp = self.app

    def async_command(self, key: str):
        """Helper function to call the process_key_event command asynchronously."""
        self.app.call_later(self.dolphie_app.key_event_manager.process_key_event, key)

    def get_command_hits(self):
        """Helper function to get all commands and format them for discovery or search."""
        commands = self.dolphie_app.command_manager.get_commands(
            self.dolphie_app.tab_manager.active_tab.dolphie.replay_file,
            self.dolphie_app.tab_manager.active_tab.dolphie.connection_source,
        )

        # Find the longest human_key length
        max_key_length = max(len(data["human_key"]) for data in commands.values())

        # Format the commands with dynamic spacing
        return {
            key: {
                # Center the human_key based on the max length and pad spaces after it
                "display": (
                    f"[$b_highlight]{data['human_key'].center(max_key_length)}[/$b_highlight]  {data['description']}"
                ),
                "text": f"{data['human_key']} {data['description']}",
                "command": partial(self.async_command, key),
                "human_key": data["human_key"],
            }
            for key, data in commands.items()
        }

    async def discover(self):
        for data in self.get_command_hits().values():
            yield DiscoveryHit(
                display=data["display"],
                text=data["text"],
                command=data["command"],
            )

    async def search(self, query: str):
        hits = []

        # Gather all hits and calculate their scores
        for data in self.get_command_hits().values():
            score = self.matcher(query).match(data["text"])
            if score > 0:
                hits.append(
                    Hit(
                        score=score,
                        match_display=data["display"],
                        text=data["text"],
                        command=data["command"],
                    )
                )

        # Sort the hits by score, descending order
        hits.sort(key=lambda hit: hit.score, reverse=True)

        for hit in hits:
            yield hit



from __future__ import annotations

import ipaddress
import os
import socket
import time
from datetime import datetime

import dolphie.DataTypes as DataTypes
import dolphie.Modules.MetricManager as MetricManager
import psutil
from dolphie.Modules.ArgumentParser import Config
from dolphie.Modules.Functions import load_host_cache_file
from dolphie.Modules.MySQL import ConnectionSource, Database
from dolphie.Modules.PerformanceSchemaMetrics import PerformanceSchemaMetrics
from dolphie.Modules.Queries import MySQLQueries
from loguru import logger
from packaging.version import parse as parse_version
from rich.text import Text
from textual.app import App
from textual.widgets import Switch


class Dolphie:
    def __init__(self, config: Config, app: App) -> None:
        self.config = config
        self.app = app
        self.app_version = config.app_version

        self.tab_id: str = None

        # Config options
        self.credential_profile = config.credential_profile
        self.user = config.user
        self.password = config.password
        self.host = config.host
        self.port = config.port
        self.socket = config.socket
        self.ssl = config.ssl
        self.host_cache_file = config.host_cache_file
        self.tab_setup_file = config.tab_setup_file
        self.refresh_interval = config.refresh_interval
        self.show_trxs_only = config.show_trxs_only
        self.show_threads_with_concurrency_tickets = False
        self.show_additional_query_columns = config.show_additional_query_columns
        self.show_statements_summary_query_digest_text_sample = False
        self.heartbeat_table = config.heartbeat_table
        self.tab_setup_available_hosts = config.tab_setup_available_hosts
        self.startup_panels = config.startup_panels
        self.graph_marker = config.graph_marker
        self.hostgroup = config.hostgroup
        self.hostgroup_hosts = config.hostgroup_hosts
        self.record_for_replay = config.record_for_replay
        self.daemon_mode = config.daemon_mode
        self.daemon_mode_panels = config.daemon_mode_panels
        self.replay_file = config.replay_file  # This denotes that we're replaying a file
        self.replay_dir = config.replay_dir
        self.replay_retention_hours = config.replay_retention_hours
        self.exclude_notify_global_vars = config.exclude_notify_global_vars

        # Set the default panels based on startup_panels to be visible
        self.panels = DataTypes.Panels()

        panels_to_show = self.daemon_mode_panels if self.daemon_mode else self.startup_panels
        for panel in panels_to_show:
            if panel in self.panels.all():
                getattr(self.panels, panel).visible = True

        self.show_idle_threads: bool = False
        self.sort_by_time_descending: bool = True

        self.reset_runtime_variables()

    def reset_runtime_variables(self):
        self.metric_manager = MetricManager.MetricManager(self.replay_file, self.daemon_mode)
        self.replica_manager = DataTypes.ReplicaManager()

        self.dolphie_start_time: datetime = datetime.now().astimezone()
        self.worker_previous_start_time: datetime = datetime.now().astimezone()
        self.worker_processing_time: float = 0
        self.polling_latency: float = 0
        self.connection_status: DataTypes.ConnectionStatus = None

        self.global_variables: dict[str, int | str] = {}
        self.global_status: dict[str, int | str] = {}
        self.binlog_status: dict[str, int | str] = {}
        self.replication_status: dict[str, int | str] = {}
        self.replication_applier_status: dict[str, list[dict[str, int | str]] | int] = {}
        self.innodb_metrics: dict[str, int | str] = {}
        self.metadata_locks: list[dict[str, int | str]] = []
        self.ddl: list[dict[str, int | str]] = []
        self.disk_io_metrics: dict[str, int | str] = {}
        self.statements_summary_metrics: dict[str, int | str] = {}
        self.system_utilization: dict[str, int | str] = {}
        self.host_cache: dict[str, str] = {}
        self.proxysql_hostgroup_summary: list[dict[str, str]] = []
        self.proxysql_mysql_query_rules: list[dict[str, str]] = []
        self.proxysql_per_second_data: dict[str, int | str] = {}
        self.proxysql_command_stats: list[dict[str, int | str]] = []
        self.processlist_threads: dict[int, DataTypes.ProcesslistThread | DataTypes.ProxySQLProcesslistThread] = {}
        self.processlist_threads_snapshot: dict[
            int, DataTypes.ProcesslistThread | DataTypes.ProxySQLProcesslistThread
        ] = {}

        # These are for group replication in replication panel
        self.is_group_replication_primary: bool = False
        self.group_replication_data: dict[str, str] = {}
        self.group_replication_members: list[dict[str, str]] = []
        self.innodb_cluster_clustersets: list[dict[str, str]] = []

        self.galera_cluster_members: list[dict[str, str]] = []

        # Filters that can be applied
        self.user_filter: str = None
        self.db_filter: str = None
        self.host_filter: str = None
        self.query_filter: str = None
        self.hostgroup_filter: int = None
        self.query_time_filter: int = None

        # Types of hosts
        self.connection_source: ConnectionSource = ConnectionSource.mysql  # mysql, proxysql
        self.connection_source_alt: ConnectionSource = ConnectionSource.mysql  # mariadb
        self.galera_cluster: bool = False
        self.group_replication: bool = False
        self.innodb_cluster: bool = False
        self.innodb_cluster_read_replica: bool = False
        self.replicaset: bool = False

        # Main connection is used for Textual's worker thread so it can run asynchronous
        db_connection_args = {
            "app": self.app,
            "host": self.host,
            "user": self.user,
            "password": self.password,
            "socket": self.socket,
            "port": self.port,
            "ssl": self.ssl,
            "auto_connect": False,
            "daemon_mode": self.daemon_mode,
        }
        self.main_db_connection = Database(**db_connection_args)
        # Secondary connection is for ad-hoc commands that are not a part of the worker thread
        self.secondary_db_connection = Database(**db_connection_args, save_connection_id=False)

        # Misc variables
        self.host_distro: str = "MySQL"
        self.host_with_port: str = f"{self.host}:{self.port}"
        self.performance_schema_enabled: bool = False
        self.use_performance_schema_for_processlist: bool = False
        self.server_uuid: str = None
        self.replication_source_uuids: set = set()
        self.host_version: str = None
        self._parsed_host_version = None
        self.pause_refresh: bool = False
        self.active_redo_logs: int = None
        self.metadata_locks_enabled: bool = False

        self.host_cache_from_file = load_host_cache_file(self.host_cache_file)

        self.file_io_data: PerformanceSchemaMetrics = None
        self.table_io_waits_data: PerformanceSchemaMetrics = None
        self.statements_summary_data: PerformanceSchemaMetrics = None

        if self.record_for_replay or self.panels.pfs_metrics.visible:
            self.pfs_metrics_last_reset_time: datetime = datetime.now().astimezone()
        else:
            # This will be set when user presses key to bring up panel
            self.pfs_metrics_last_reset_time: datetime = None

        self.update_switches_after_reset()

        try:
            # Get the IP address of the monitored host
            monitored_ip = socket.gethostbyname(self.host)

            # Enable system metrics if using a socket file or if monitored host is localhost
            if self.socket or monitored_ip == "127.0.0.1" or monitored_ip == socket.gethostbyname(socket.gethostname()):
                self.enable_system_utilization = True
            else:
                self.enable_system_utilization = False
        except socket.gaierror:
            self.enable_system_utilization = False

    def db_connect(self):
        self.main_db_connection.connect()
        if not self.daemon_mode:
            self.secondary_db_connection.connect()

        self.connection_source = self.main_db_connection.source
        self.connection_source_alt = self.connection_source
        if self.connection_source == ConnectionSource.proxysql:
            self.host_distro = ConnectionSource.proxysql
            self.host_with_port = f"{self.host}:{self.port}"

        self.metric_manager.connection_source = self.connection_source

        # Add host to tab setup file if it doesn't exist
        self.add_host_to_tab_setup_file()

    def configure_mysql_variables(self):
        global_variables = self.global_variables

        # Galera cluster check
        self.galera_cluster = global_variables.get("wsrep_on") == "ON" or bool(
            global_variables.get("wsrep_cluster_address")
        )

        self.host_distro, self.connection_source_alt = self.determine_distro_and_connection_source_alt(global_variables)

        # For RDS and Azure, we will use the host specified to connect with since hostname isn't related to the endpoint
        if ".rds.amazonaws.com" in self.host:
            self.host_with_port = f"{self.host.split('.rds.amazonaws.com')[0]}:{self.port}"
        elif ".mysql.database.azure.com" in self.host:
            self.host_with_port = f"{self.host.split('.mysql.database.azure.com')[0]}:{self.port}"
        else:
            mysql_hostname = global_variables.get("hostname")
            self.host_with_port = f"{mysql_hostname}:{self.port}" if mysql_hostname else f"{self.host}:{self.port}"

        # Server UUID configuration (mainly for replication & errant transactions)
        self.server_uuid = global_variables.get("server_uuid")
        if self.connection_source_alt == ConnectionSource.mariadb and self.is_mysql_version_at_least("10.0"):
            self.server_uuid = global_variables.get("server_id")

        # Performance schema
        self.performance_schema_enabled = global_variables.get("performance_schema") == "ON"
        self.use_performance_schema_for_processlist = self.performance_schema_enabled

        # Cluster type detection
        cluster_type = self.group_replication_data.get("cluster_type")
        if cluster_type == "ar":
            self.replicaset = True
        elif cluster_type == "gr":
            self.innodb_cluster = True
            if self.group_replication_data.get("instance_type") == "read-replica":
                self.innodb_cluster = False  # Not a part of the cluster if it's a read replica
                self.innodb_cluster_read_replica = True

        # Group replication host check
        if not self.innodb_cluster and global_variables.get("group_replication_group_name"):
            self.group_replication = True

    def determine_distro_and_connection_source_alt(
        self, global_variables: dict[str, int | str]
    ) -> tuple[str, ConnectionSource]:
        is_percona = "percona" in global_variables.get("version_comment", "").casefold()
        is_mariadb = any(variable.startswith("aria_") for variable in global_variables)
        is_rds = "rdsdb" in self.global_variables.get("basedir", "").casefold()
        is_aurora = self.global_variables.get("aurora_version")
        is_azure = self.global_variables.get("aad_auth_only")
        is_galera_cluster = self.galera_cluster

        # MariaDB
        if is_mariadb:
            if is_rds:
                return "Amazon RDS (MariaDB)", ConnectionSource.mariadb
            if is_azure:
                return "Azure MariaDB", ConnectionSource.mariadb
            if is_galera_cluster:
                return "MariaDB", ConnectionSource.mariadb
            return "MariaDB", ConnectionSource.mariadb

        # Percona
        if is_percona:
            if is_galera_cluster:
                return "Percona Server", ConnectionSource.mysql
            return "Percona Server", ConnectionSource.mysql

        # Standard MySQL
        if is_aurora:
            return "Amazon Aurora", ConnectionSource.mysql
        if is_rds:
            return "Amazon RDS (MySQL)", ConnectionSource.mysql
        if is_azure:
            return "Azure MySQL", ConnectionSource.mysql

        return "MySQL", ConnectionSource.mysql

    def build_kill_query(self, thread_id: int) -> str:
        is_rds = "rdsdb" in self.global_variables.get("basedir", "").casefold()
        is_aurora = self.global_variables.get("aurora_version")
        is_azure = self.global_variables.get("aad_auth_only")

        if is_rds or is_aurora:
            return f"CALL mysql.rds_kill({thread_id})"
        if is_azure:
            return f"CALL mysql.az_kill({thread_id})"
        if self.connection_source == ConnectionSource.proxysql:
            return f"KILL CONNECTION {thread_id}"

        return f"KILL {thread_id}"

    def collect_system_utilization(self):
        if not self.enable_system_utilization:
            return

        virtual_memory = psutil.virtual_memory()
        swap_memory = psutil.swap_memory()
        network_io = psutil.net_io_counters()
        disk_io = psutil.disk_io_counters()

        self.system_utilization = {
            "Uptime": int(time.time() - psutil.boot_time()),
            "CPU_Count": psutil.cpu_count(logical=True),
            "CPU_Percent": psutil.cpu_percent(interval=0),
            "Memory_Total": virtual_memory.total,
            "Memory_Used": virtual_memory.used,
            "Swap_Total": swap_memory.total,
            "Swap_Used": swap_memory.used,
            "Network_Up": network_io.bytes_sent,
            "Network_Down": network_io.bytes_recv,
            "Disk_Read": disk_io.read_count,
            "Disk_Write": disk_io.write_count,
        }

        # Include the load average if it's available
        try:
            self.system_utilization["CPU_Load_Avg"] = psutil.getloadavg()  # 1, 5, and 15 minute load averages
        except AttributeError:
            pass

    def get_group_replication_metadata(self):
        # Check to get information on what cluster/instance type it is
        if self.is_mysql_version_at_least("8.1"):
            query = MySQLQueries.determine_cluster_type_81
        else:
            query = MySQLQueries.determine_cluster_type_8

        self.main_db_connection.execute(query, ignore_error=True)
        self.group_replication_data = self.main_db_connection.fetchone()

    def add_host_to_tab_setup_file(self):
        if self.daemon_mode:
            return

        with open(self.tab_setup_file, "a+") as file:
            file.seek(0)
            lines = file.readlines()

            host = f"{self.host}:{self.port}\n" if self.port != 3306 else f"{self.host}\n"

            if host not in lines:
                file.write(host)
                self.tab_setup_available_hosts.append(host[:-1])  # remove the \n

    def is_mysql_version_at_least(self, target: str, use_version: str = None):
        if use_version:
            parsed_source = parse_version(use_version)
        else:
            parsed_source = self._parsed_host_version

        return parsed_source >= parse_version(target)

    def parse_server_version(self, version: str) -> str:
        if not version:
            return "N/A"

        major, minor, patch = version.split(".", 2)
        patch = patch.split("-", 1)[0]

        parsed = f"{major}.{minor}.{patch}"
        self._parsed_host_version = parse_version(parsed)

        return parsed

    def get_hostname(self, host):
        if host in self.host_cache:
            return self.host_cache[host]

        if self.host_cache_from_file and host in self.host_cache_from_file:
            self.host_cache[host] = self.host_cache_from_file[host]
            return self.host_cache_from_file[host]

        try:
            ipaddress.IPv4Network(host)
            hostname = socket.gethostbyaddr(host)[0]
            self.host_cache[host] = hostname
        except (OSError, ValueError):
            self.host_cache[host] = host
            hostname = host

        return hostname

    def update_switches_after_reset(self):
        # Set the graph switches to what they're currently selected to after a reset
        switches = self.app.query(f".switch_container_{self.tab_id} Switch")
        for switch in switches:
            switch: Switch
            metric_instance_name = switch.name
            metric = switch.id

            metric_instance = getattr(self.metric_manager.metrics, metric_instance_name)
            metric_data: MetricManager.MetricData = getattr(metric_instance, metric)
            metric_data.visible = switch.value

    def determine_proxysql_refresh_interval(self) -> float:
        # If we have a lot of client connections, increase the refresh interval based on the
        # proxysql process execution time. René asked for this to be added to reduce load on ProxySQL
        client_connections = self.global_status.get("Client_Connections_connected", 0)
        if client_connections > 30000:
            percentage = 0.60
        elif client_connections > 20000:
            percentage = 0.50
        elif client_connections > 10000:
            percentage = 0.40
        else:
            percentage = 0

        if percentage:
            refresh_interval = self.refresh_interval + (self.worker_processing_time * percentage)
        else:
            refresh_interval = self.refresh_interval

        return refresh_interval

    def validate_metadata_locks_enabled(self):
        if not self.is_mysql_version_at_least("5.7") or not self.performance_schema_enabled:
            logger.warning(
                "Metadata Locks requires MySQL 5.7+ with Performance Schema enabled - will not capture that data"
            )
            return

        query = """
            SELECT enabled FROM performance_schema.setup_instruments WHERE name = 'wait/lock/metadata/sql/mdl'
        """
        self.main_db_connection.execute(query)
        row = self.main_db_connection.fetchone()
        if row and row.get("enabled") == "NO":
            logger.warning(
                "Metadata Locks requires Performance Schema to have"
                " wait/lock/metadata/sql/mdl enabled in setup_instruments table - will not capture that data"
            )
            return

        self.metadata_locks_enabled = True

    def get_replay_files(self):
        """Gets a list of replay files in the replay directory.

        Returns:
            list: A list of tuples in the format (full_path, formatted host name + replay name).
        """
        if not self.replay_dir or not os.path.exists(self.replay_dir):
            return []

        replay_files = []
        try:
            with os.scandir(self.replay_dir) as entries:
                for entry in entries:
                    if entry.is_dir():
                        entry_path = entry.path
                        for file in os.scandir(entry_path):
                            if file.is_file():
                                # Get first 30 characters of the host name
                                host_name = entry.name[:30]

                                # Only set port if the host name is 30 characters or more
                                port = ""
                                if len(entry.name) >= 30 and "_" in entry.name:
                                    port = "_" + entry.name.rsplit("_", 1)[-1]

                                formatted_replay_name = f"[label]{host_name}{port}[/label]"
                                formatted_replay_name += f": [b light_blue]{file.name}[/b light_blue]"

                                replay_files.append((file.path, Text.from_markup(formatted_replay_name)))
        except OSError as e:
            self.app.notify(str(e), title="Error getting replay files", severity="error")

        # Sort replay_files by the file path
        replay_files.sort(key=lambda x: x[0])

        return replay_files

    def reset_pfs_metrics_deltas(self, reset_fully: bool = False):
        for instance in [self.file_io_data, self.table_io_waits_data]:
            if not instance:
                continue

            if reset_fully:
                instance.internal_data = {}
                instance.filtered_data = {}
            else:
                for data in (instance.internal_data, instance.filtered_data):
                    for file_data in data.values():
                        for metric_data in file_data.get("metrics", file_data).values():
                            # For filtered_data so replay data is smaller in size
                            if "d" in metric_data:
                                metric_data["d"] = 0
                            elif "delta" in metric_data:
                                metric_data["delta"] = 0

        self.pfs_metrics_last_reset_time = datetime.now().astimezone()
